import json

import boto3
import pytest
from botocore.exceptions import ClientError

from moto import mock_aws


@mock_aws
def test_create_and_get_replication_task():
    client = boto3.client("dms", region_name="us-east-1")

    client.create_replication_task(
        ReplicationTaskIdentifier="test",
        SourceEndpointArn="source-endpoint-arn",
        TargetEndpointArn="target-endpoint-arn",
        ReplicationInstanceArn="replication-instance-arn",
        MigrationType="full-load",
        TableMappings='{"rules":[]}',
        ReplicationTaskSettings='{"Logging":{} }',
    )

    tasks = client.describe_replication_tasks(
        Filters=[{"Name": "replication-task-id", "Values": ["test"]}]
    )

    assert len(tasks["ReplicationTasks"]) == 1
    task = tasks["ReplicationTasks"][0]
    assert task["ReplicationTaskIdentifier"] == "test"
    assert task["SourceEndpointArn"] == "source-endpoint-arn"
    assert task["TargetEndpointArn"] == "target-endpoint-arn"
    assert task["ReplicationInstanceArn"] == "replication-instance-arn"
    assert task["MigrationType"] == "full-load"
    assert task["Status"] == "creating"
    assert task["TableMappings"] == '{"rules":[]}'
    assert isinstance(json.loads(task["TableMappings"]), dict)

    assert task["ReplicationTaskSettings"] == '{"Logging":{} }'
    assert isinstance(json.loads(task["ReplicationTaskSettings"]), dict)


@mock_aws
def test_create_existing_replication_task_throws_error():
    client = boto3.client("dms", region_name="us-east-1")

    client.create_replication_task(
        ReplicationTaskIdentifier="test",
        SourceEndpointArn="source-endpoint-arn",
        TargetEndpointArn="target-endpoint-arn",
        ReplicationInstanceArn="replication-instance-arn",
        MigrationType="full-load",
        TableMappings='{"rules":[]}',
    )

    with pytest.raises(ClientError) as ex:
        client.create_replication_task(
            ReplicationTaskIdentifier="test",
            SourceEndpointArn="source-endpoint-arn",
            TargetEndpointArn="target-endpoint-arn",
            ReplicationInstanceArn="replication-instance-arn",
            MigrationType="full-load",
            TableMappings='{"rules":[]}',
        )

    assert ex.value.operation_name == "CreateReplicationTask"
    assert ex.value.response["Error"]["Code"] == "ResourceAlreadyExistsFault"
    assert (
        ex.value.response["Error"]["Message"]
        == "The resource you are attempting to create already exists."
    )


@mock_aws
def test_start_replication_task():
    client = boto3.client("dms", region_name="us-east-1")

    response = client.create_replication_task(
        ReplicationTaskIdentifier="test",
        SourceEndpointArn="source-endpoint-arn",
        TargetEndpointArn="target-endpoint-arn",
        ReplicationInstanceArn="replication-instance-arn",
        MigrationType="full-load",
        TableMappings='{"rules":[]}',
    )
    task_arn = response["ReplicationTask"]["ReplicationTaskArn"]
    client.start_replication_task(
        ReplicationTaskArn=task_arn, StartReplicationTaskType="start-replication"
    )
    tasks = client.describe_replication_tasks(
        Filters=[{"Name": "replication-task-arn", "Values": [task_arn]}]
    )

    assert tasks["ReplicationTasks"][0]["Status"] == "running"


@mock_aws
def test_start_replication_task_throws_resource_not_found_error():
    client = boto3.client("dms", region_name="us-east-1")

    with pytest.raises(ClientError) as ex:
        client.start_replication_task(
            ReplicationTaskArn="does-not-exist",
            StartReplicationTaskType="start-replication",
        )

    assert ex.value.operation_name == "StartReplicationTask"
    assert ex.value.response["Error"]["Code"] == "ResourceNotFoundFault"
    assert (
        ex.value.response["Error"]["Message"] == "Replication task could not be found."
    )


@mock_aws
def test_stop_replication_task_throws_invalid_state_error():
    client = boto3.client("dms", region_name="us-east-1")

    response = client.create_replication_task(
        ReplicationTaskIdentifier="test",
        SourceEndpointArn="source-endpoint-arn",
        TargetEndpointArn="target-endpoint-arn",
        ReplicationInstanceArn="replication-instance-arn",
        MigrationType="full-load",
        TableMappings='{"rules":[]}',
    )
    task_arn = response["ReplicationTask"]["ReplicationTaskArn"]

    with pytest.raises(ClientError) as ex:
        client.stop_replication_task(ReplicationTaskArn=task_arn)

    assert ex.value.operation_name == "StopReplicationTask"
    assert ex.value.response["Error"]["Code"] == "InvalidResourceStateFault"
    assert ex.value.response["Error"]["Message"] == "Replication task is not running"


@mock_aws
def test_stop_replication_task_throws_resource_not_found_error():
    client = boto3.client("dms", region_name="us-east-1")

    with pytest.raises(ClientError) as ex:
        client.stop_replication_task(ReplicationTaskArn="does-not-exist")

    assert ex.value.operation_name == "StopReplicationTask"
    assert ex.value.response["Error"]["Code"] == "ResourceNotFoundFault"
    assert (
        ex.value.response["Error"]["Message"] == "Replication task could not be found."
    )


@mock_aws
def test_stop_replication_task():
    client = boto3.client("dms", region_name="us-east-1")

    response = client.create_replication_task(
        ReplicationTaskIdentifier="test",
        SourceEndpointArn="source-endpoint-arn",
        TargetEndpointArn="target-endpoint-arn",
        ReplicationInstanceArn="replication-instance-arn",
        MigrationType="full-load",
        TableMappings='{"rules":[]}',
    )
    task_arn = response["ReplicationTask"]["ReplicationTaskArn"]
    client.start_replication_task(
        ReplicationTaskArn=task_arn, StartReplicationTaskType="start-replication"
    )
    client.stop_replication_task(ReplicationTaskArn=task_arn)
    tasks = client.describe_replication_tasks(
        Filters=[{"Name": "replication-task-arn", "Values": [task_arn]}]
    )

    assert tasks["ReplicationTasks"][0]["Status"] == "stopped"


@mock_aws
def test_delete_replication_task():
    client = boto3.client("dms", region_name="us-east-1")

    response = client.create_replication_task(
        ReplicationTaskIdentifier="test",
        SourceEndpointArn="source-endpoint-arn",
        TargetEndpointArn="target-endpoint-arn",
        ReplicationInstanceArn="replication-instance-arn",
        MigrationType="full-load",
        TableMappings='{"rules":[]}',
    )
    task_arn = response["ReplicationTask"]["ReplicationTaskArn"]
    client.delete_replication_task(ReplicationTaskArn=task_arn)
    tasks = client.describe_replication_tasks(
        Filters=[{"Name": "replication-task-arn", "Values": [task_arn]}]
    )

    assert len(tasks["ReplicationTasks"]) == 0


@mock_aws
def test_delete_replication_task_throws_resource_not_found_error():
    client = boto3.client("dms", region_name="us-east-1")

    with pytest.raises(ClientError) as ex:
        client.delete_replication_task(ReplicationTaskArn="does-not-exist")

    assert ex.value.operation_name == "DeleteReplicationTask"
    assert ex.value.response["Error"]["Code"] == "ResourceNotFoundFault"
    assert (
        ex.value.response["Error"]["Message"] == "Replication task could not be found."
    )


@mock_aws
def test_create_replication_instance():
    client = boto3.client("dms", region_name="us-east-1")

    response = client.create_replication_instance(
        ReplicationInstanceIdentifier="test-instance",
        ReplicationInstanceClass="dms.t2.micro",
        AllocatedStorage=50,
        VpcSecurityGroupIds=["sg-12345"],
        AvailabilityZone="us-east-1a",
        ReplicationSubnetGroupIdentifier="default-subnet-group",
        PreferredMaintenanceWindow="sun:06:00-sun:14:00",
        MultiAZ=False,
        EngineVersion="3.4.6",
        AutoMinorVersionUpgrade=True,
        Tags=[{"Key": "Name", "Value": "Test Instance"}],
        PubliclyAccessible=True,
        NetworkType="IPV4",
    )

    instance = response["ReplicationInstance"]
    assert instance["ReplicationInstanceIdentifier"] == "test-instance"
    assert instance["ReplicationInstanceClass"] == "dms.t2.micro"
    assert instance["AllocatedStorage"] == 50
    assert instance["AvailabilityZone"] == "us-east-1a"
    assert instance["ReplicationInstanceStatus"] == "creating"
    assert instance["MultiAZ"] is False
    assert instance["EngineVersion"] == "3.4.6"
    assert instance["AutoMinorVersionUpgrade"] is True
    assert instance["PubliclyAccessible"] is True
    assert instance["NetworkType"] == "IPV4"

    arn = instance["ReplicationInstanceArn"]
    assert arn.startswith("arn:aws:dms:us-east-1:")
    assert ":rep:test-instance" in arn

    response = client.describe_replication_instances(
        Filters=[{"Name": "replication-instance-id", "Values": ["test-instance"]}]
    )
    assert len(response["ReplicationInstances"]) == 1
    assert (
        response["ReplicationInstances"][0]["ReplicationInstanceIdentifier"]
        == "test-instance"
    )


@mock_aws
def test_describe_replication_instances():
    client = boto3.client("dms", region_name="us-east-1")

    client.create_replication_instance(
        ReplicationInstanceIdentifier="test-instance-1",
        ReplicationInstanceClass="dms.t2.micro",
        EngineVersion="3.4.5",
    )

    client.create_replication_instance(
        ReplicationInstanceIdentifier="test-instance-2",
        ReplicationInstanceClass="dms.t2.small",
        EngineVersion="3.4.6",
    )

    response = client.describe_replication_instances()
    instances = response["ReplicationInstances"]
    assert len(instances) == 2

    response = client.describe_replication_instances(
        Filters=[{"Name": "replication-instance-class", "Values": ["dms.t2.micro"]}]
    )
    instances = response["ReplicationInstances"]
    assert len(instances) == 1
    assert instances[0]["ReplicationInstanceIdentifier"] == "test-instance-1"

    response = client.describe_replication_instances(
        Filters=[{"Name": "engine-version", "Values": ["3.4.6"]}]
    )
    instances = response["ReplicationInstances"]
    assert len(instances) == 1
    assert instances[0]["ReplicationInstanceIdentifier"] == "test-instance-2"


@mock_aws
def test_create_endpoint():
    region_name = "ap-southeast-1"
    client = boto3.client("dms", region_name=region_name)

    response = client.create_endpoint(
        EndpointIdentifier="test-endpoint",
        EndpointType="source",
        EngineName="mysql",
        Username="admin",
        ServerName="test-server",
        Port=3306,
        DatabaseName="test-db",
        ExtraConnectionAttributes="",
        KmsKeyId="test-kms-key-id",
        CertificateArn="test-certificate-arn",
        SslMode="require",
        ServiceAccessRoleArn="test-service-role-arn",
        ExternalTableDefinition="",
    )

    endpoint = response["Endpoint"]

    # Assertions for the required fields
    assert endpoint["EndpointIdentifier"] == "test-endpoint"
    assert endpoint["EndpointType"] == "source"
    assert endpoint["EngineName"] == "mysql"
    assert endpoint["Username"] == "admin"
    assert endpoint["ServerName"] == "test-server"
    assert endpoint["Port"] == 3306
    assert endpoint["DatabaseName"] == "test-db"
    assert endpoint["Status"] == "creating"
    assert endpoint["KmsKeyId"] == "test-kms-key-id"
    assert (
        f"arn:aws:dms:{region_name}:123456789012:endpoint:" in endpoint["EndpointArn"]
    )
    assert endpoint["SslMode"] == "require"


@mock_aws
def test_create_endpoint_resource_identifier():
    client = boto3.client("dms", region_name="ap-southeast-1")

    resource_identifier = "test-resource-identifier"
    response = client.create_endpoint(
        EndpointIdentifier="test-endpoint",
        EndpointType="source",
        EngineName="mysql",
        ResourceIdentifier=resource_identifier,
    )

    endpoint = response["Endpoint"]

    # Assertions for the required fields
    assert endpoint["EndpointIdentifier"] == "test-endpoint"
    assert endpoint["EndpointType"] == "source"
    assert endpoint["EngineName"] == "mysql"
    assert resource_identifier == endpoint["EndpointArn"].split(":")[-1]


@mock_aws
def test_describe_endpoints():
    client = boto3.client("dms", region_name="ap-southeast-1")
    num_endpoints = 3
    for i in range(num_endpoints):
        client.create_endpoint(
            EndpointIdentifier=f"test-endpoint-{i}",
            EndpointType="source",
            EngineName="mysql",
        )

    resp = client.describe_endpoints()
    assert len(resp["Endpoints"]) == num_endpoints


@mock_aws
def test_describe_endpoints_filter():
    client = boto3.client("dms", region_name="ap-southeast-1")

    for i in range(3):
        client.create_endpoint(
            EndpointIdentifier=f"test-endpoint-{i}",
            EndpointType="source",
            EngineName="mysql",
        )

    endpoint_filter = {"Name": "endpoint-id", "Values": ["test-endpoint-1"]}
    resp = client.describe_endpoints(Filters=[endpoint_filter])
    assert len(resp["Endpoints"]) == 1

    engine_filter = {"Name": "engine-name", "Values": ["mysql"]}
    resp = client.describe_endpoints(Filters=[engine_filter])
    assert len(resp["Endpoints"]) == 3


@mock_aws
def test_list_tags_for_resource_replication_instance():
    client = boto3.client("dms", region_name="eu-west-1")

    response = client.create_replication_instance(
        ReplicationInstanceIdentifier="test",
        ReplicationInstanceClass="dms.c4.large",
        Tags=[{"Key": "Environment", "Value": "Production"}],
    )

    arn = response["ReplicationInstance"]["ReplicationInstanceArn"]

    resp = client.list_tags_for_resource(ResourceArn=arn)
    assert resp["TagList"] == [
        {"Key": "Environment", "Value": "Production", "ResourceArn": arn}
    ]


@mock_aws
def test_list_tags_for_resource_endpoint():
    client = boto3.client("dms", region_name="eu-west-1")

    response = client.create_endpoint(
        EndpointIdentifier="test-endpoint",
        EndpointType="source",
        EngineName="mysql",
        Tags=[{"Key": "Name", "Value": "Test Endpoint"}],
    )
    arn = response["Endpoint"]["EndpointArn"]

    resp = client.list_tags_for_resource(ResourceArn=arn)
    assert resp["TagList"] == [
        {"Key": "Name", "Value": "Test Endpoint", "ResourceArn": arn}
    ]


@mock_aws
def test_list_tags_for_resource_endpoints():
    client = boto3.client("dms", region_name="eu-west-1")

    response = client.create_endpoint(
        EndpointIdentifier="test-endpoint",
        EndpointType="source",
        EngineName="mysql",
        Tags=[{"Key": "Name", "Value": "Test Endpoint"}],
    )
    endpoint_arn1 = response["Endpoint"]["EndpointArn"]

    response = client.create_endpoint(
        EndpointIdentifier="test-endpoint",
        EndpointType="source",
        EngineName="mysql",
        Tags=[{"Key": "Name", "Value": "Test Endpoint"}],
    )
    endpoint_arn2 = response["Endpoint"]["EndpointArn"]

    resp = client.list_tags_for_resource(ResourceArnList=[endpoint_arn1, endpoint_arn2])
    assert len(resp["TagList"]) == 2



from contextlib import nullcontext
from functools import wraps
from uuid import uuid4

import boto3

from moto import mock_aws
from tests import allow_aws_request


def ec2_aws_verified(
    create_vpc: bool = False,
    create_sg: bool = False,
):
    """
    Function that is verified to work against AWS.
    Can be run against AWS at any time by setting:
      MOTO_TEST_ALLOW_AWS_REQUEST=true

    If this environment variable is not set, the function runs in a `mock_aws` context.

    """

    def inner(func):
        @wraps(func)
        def pagination_wrapper(**kwargs):
            context = nullcontext() if allow_aws_request() else mock_aws()

            with context:
                return _invoke_func(create_vpc, create_sg, func=func, kwargs=kwargs)

        return pagination_wrapper

    return inner


def _invoke_func(create_vpc: bool, create_sg: bool, func, kwargs):
    ec2_client = boto3.client("ec2", "us-east-1")
    kwargs["ec2_client"] = ec2_client

    vpc_id = sg_id = None
    if create_vpc:
        vpc = ec2_client.create_vpc(CidrBlock="10.0.0.0/24")
        vpc_id = vpc["Vpc"]["VpcId"]
        kwargs["vpc_id"] = vpc_id

    if create_sg:
        sg_name = f"test_{str(uuid4())[0:6]}"
        sg = ec2_client.create_security_group(
            Description="test", GroupName=sg_name, VpcId=vpc_id
        )
        sg_id = sg["GroupId"]
        kwargs["sg_id"] = sg_id

    try:
        func(**kwargs)
    finally:
        if sg_id:
            ec2_client.delete_security_group(GroupId=sg_id)
        if vpc_id:
            ec2_client.delete_vpc(VpcId=vpc_id)



from moto.core.responses import ActionResult

from ._base_response import EC2BaseResponse


class TransitGateways(EC2BaseResponse):
    def create_transit_gateway(self) -> ActionResult:
        description = self._get_param("Description") or None
        options = self._get_multi_param_dict("Options")
        tags = self._get_multi_param("TagSpecification")
        if tags:
            tags = tags[0].get("Tag")

        transit_gateway = self.ec2_backend.create_transit_gateway(
            description=description, options=options, tags=tags
        )

        # creating default route table
        transit_gateway_route_table = (
            self.ec2_backend.create_transit_gateway_route_table(
                transit_gateway_id=transit_gateway.id,
                tags={},
                default_association_route_table=True,
                default_propagation_route_table=True,
            )
        )
        transit_gateway.options["AssociationDefaultRouteTableId"] = (
            transit_gateway_route_table.id
        )
        transit_gateway.options["PropagationDefaultRouteTableId"] = (
            transit_gateway_route_table.id
        )
        return ActionResult({"TransitGateway": transit_gateway})

    def delete_transit_gateway(self) -> ActionResult:
        transit_gateway_id = self._get_param("TransitGatewayId")
        transit_gateway = self.ec2_backend.delete_transit_gateway(transit_gateway_id)
        return ActionResult({"TransitGateway": transit_gateway})

    def describe_transit_gateways(self) -> ActionResult:
        transit_gateway_ids = self._get_multi_param("TransitGatewayIds")
        filters = self._filters_from_querystring()
        transit_gateways = self.ec2_backend.describe_transit_gateways(
            filters, transit_gateway_ids
        )
        return ActionResult({"TransitGateways": transit_gateways})

    def modify_transit_gateway(self) -> ActionResult:
        transit_gateway_id = self._get_param("TransitGatewayId")
        description = self._get_param("Description") or None
        options = self._get_multi_param_dict("Options")
        transit_gateway = self.ec2_backend.modify_transit_gateway(
            transit_gateway_id=transit_gateway_id,
            description=description,
            options=options,
        )
        return ActionResult({"TransitGateway": transit_gateway})



import boto3
import pytest
from botocore.exceptions import ClientError

from moto import mock_aws


@mock_aws
def test_associate_resource_cloudformation():
    client = boto3.client("servicecatalog-appregistry", region_name="us-east-1")
    stack_name = "foo"
    stack_id = boto3.client("cloudformation", region_name="us-east-1").create_stack(
        StackName=stack_name,
        TemplateBody="""
---
AWSTemplateFormatVersion: '2010-09-09'
Resources:
  S3:
    Type: AWS::S3::Bucket
    Properties:
      BucketName: foo
""",
    )["StackId"]
    create = client.create_application(name="testapp", description="blah")
    resp = client.associate_resource(
        application=create["application"]["id"],
        resource=stack_id,
        resourceType="CFN_STACK",
        options=["APPLY_APPLICATION_TAG"],
    )

    assert "applicationArn" in resp
    assert "resourceArn" in resp
    assert "options" in resp
    assert resp["applicationArn"] == create["application"]["arn"]


@mock_aws
def test_associate_resource_cloudformation_validation_error():
    client = boto3.client("servicecatalog-appregistry", region_name="us-east-1")
    create = client.create_application(name="testapp", description="blah")
    stack_name = "foo"
    with pytest.raises(ClientError) as exc:
        client.associate_resource(
            application=create["application"]["id"],
            resource=stack_name,
            resourceType="CFN_STACK",
            options=["APPLY_APPLICATION_TAG"],
        )
    err = exc.value.response["Error"]
    assert err["Code"] == "ResourceNotFoundException"
    assert err["Message"] == f"No CloudFormation stack called '{stack_name}' found"



from moto.core.exceptions import JsonRESTError


class BadRequestException(JsonRESTError):
    code = 400

    def __init__(self, message: str):
        super().__init__(error_type="BadRequestException", message=message)


class NotFoundException(JsonRESTError):
    code = 404

    def __init__(
        self, message: str = "One or more of the specified resources don't exist."
    ):
        super().__init__(error_type="NotFoundException", message=message)



"""ServiceCatalogBackend class with methods for supported APIs."""

import uuid
from datetime import datetime
from typing import Any, Dict, List, Optional, Tuple

from moto.core.base_backend import BackendDict, BaseBackend
from moto.core.common_models import BaseModel
from moto.utilities.paginator import paginate
from moto.utilities.utils import get_partition

PAGINATION_MODEL = {
    "list_portfolio_access": {
        "input_token": "page_token",
        "limit_key": "page_size",
        "limit_default": 20,
        "unique_attribute": "account_id",
    }
}


class Portfolio(BaseModel):
    """Portfolio resource."""

    def __init__(
        self,
        portfolio_id: str,
        display_name: str,
        description: str,
        provider_name: str,
        tags: List[Dict[str, str]],
        region_name: str,
        account_id: str,
    ) -> None:
        self.id = portfolio_id
        self.display_name = display_name
        self.description = description
        self.provider_name = provider_name
        self.created_time = datetime.now()
        self.tags = tags
        self.region_name = region_name
        self.account_id = account_id

    @property
    def arn(self) -> str:
        return f"arn:{get_partition(self.region_name)}:catalog:{self.region_name}:{self.account_id}:portfolio/{self.id}"

    def to_dict(self) -> Dict[str, str]:
        return {
            "Id": self.id,
            "ARN": self.arn,
            "DisplayName": self.display_name,
            "Description": self.description,
            "CreatedTime": self.created_time.isoformat(),
            "ProviderName": self.provider_name,
        }


class ServiceCatalogBackend(BaseBackend):
    """Implementation of ServiceCatalog APIs."""

    def __init__(self, region_name: str, account_id: str) -> None:
        super().__init__(region_name, account_id)
        self.portfolio_access: Dict[str, List[str]] = {}
        self.portfolios: Dict[str, Portfolio] = {}
        self.idempotency_tokens: Dict[str, str] = {}
        self.portfolio_share_tokens: Dict[str, List[str]] = {}

    @paginate(pagination_model=PAGINATION_MODEL)
    def list_portfolio_access(
        self,
        accept_language: Optional[str],
        portfolio_id: str,
        organization_parent_id: Optional[str],
        page_token: Optional[str],
        page_size: Optional[int] = None,
    ) -> List[Dict[str, str]]:
        # TODO: Implement organization_parent_id and accept_language

        account_ids = self.portfolio_access.get(portfolio_id, [])
        return [{"account_id": account_id} for account_id in account_ids]

    def delete_portfolio(self, accept_language: Optional[str], id: str) -> None:
        # TODO: Implement accept_language

        if id in self.portfolio_access:
            del self.portfolio_access[id]

        if id in self.portfolios:
            del self.portfolios[id]

        return None

    def delete_portfolio_share(
        self,
        accept_language: Optional[str],
        portfolio_id: str,
        account_id: Optional[str],
        organization_node: Optional[Dict[str, str]],
    ) -> Optional[str]:
        # TODO: Implement accept_language

        if (
            portfolio_id in self.portfolio_access
            and account_id in self.portfolio_access[portfolio_id]
        ):
            self.portfolio_access[portfolio_id].remove(account_id)

        portfolio_share_token = None
        if organization_node:
            org_type = organization_node.get("Type", "")
            org_value = organization_node.get("Value", "")

            # Arbitrary naming for the portfolio share token
            portfolio_share_token = f"share-{portfolio_id}-{org_type}-{org_value}"

            if portfolio_id in self.portfolio_share_tokens:
                tokens_to_remove = []

                for token in self.portfolio_share_tokens[portfolio_id]:
                    if org_type in token and org_value in token:
                        tokens_to_remove.append(token)

                for token in tokens_to_remove:
                    if token in self.portfolio_share_tokens[portfolio_id]:
                        self.portfolio_share_tokens[portfolio_id].remove(token)

        return portfolio_share_token

    def create_portfolio(
        self,
        accept_language: Optional[str],
        display_name: str,
        description: Optional[str],
        provider_name: str,
        tags: Optional[List[Dict[str, str]]],
        idempotency_token: Optional[str],
    ) -> Tuple[Dict[str, str], List[Dict[str, str]]]:
        # TODO: Implement accept_language

        if idempotency_token and idempotency_token in self.idempotency_tokens:
            portfolio_id = self.idempotency_tokens[idempotency_token]
            portfolio = self.portfolios[portfolio_id]
            return portfolio.to_dict(), portfolio.tags

        portfolio_id = str(uuid.uuid4())

        portfolio = Portfolio(
            portfolio_id=portfolio_id,
            display_name=display_name,
            description=description or "",
            provider_name=provider_name,
            tags=tags or [],
            region_name=self.region_name,
            account_id=self.account_id,
        )

        self.portfolios[portfolio_id] = portfolio

        if idempotency_token:
            self.idempotency_tokens[idempotency_token] = portfolio_id

        self.portfolio_access[portfolio_id] = []

        return portfolio.to_dict(), portfolio.tags

    def create_portfolio_share(
        self,
        accept_language: Optional[str],
        portfolio_id: str,
        account_id: Optional[str],
        organization_node: Optional[Dict[str, str]],
        share_tag_options: bool,
        share_principals: bool,
    ) -> Optional[str]:
        # TODO: Implement accept_language

        if portfolio_id not in self.portfolios:
            return None

        if account_id:
            if portfolio_id not in self.portfolio_access:
                self.portfolio_access[portfolio_id] = []

            if account_id not in self.portfolio_access[portfolio_id]:
                self.portfolio_access[portfolio_id].append(account_id)

            return None

        portfolio_share_token = None
        if organization_node:
            org_type = organization_node.get("Type", "")
            org_value = organization_node.get("Value", "")

            # Arbitrary org for the portfolio share token
            portfolio_share_token = f"share-{portfolio_id}-{org_type}-{org_value}"

            if share_tag_options:
                portfolio_share_token += "-tags"
            if share_principals:
                portfolio_share_token += "-principals"

            if portfolio_id not in self.portfolio_share_tokens:
                self.portfolio_share_tokens[portfolio_id] = []

            if portfolio_share_token not in self.portfolio_share_tokens[portfolio_id]:
                self.portfolio_share_tokens[portfolio_id].append(portfolio_share_token)

        return portfolio_share_token

    def list_portfolios(
        self,
        accept_language: Optional[str] = None,
        page_token: Optional[str] = None,
        page_size: Optional[int] = None,
    ) -> Tuple[List[Dict[str, str]], Optional[str]]:
        """TODO: Implement pagination and accept_language"""
        portfolio_details = [
            portfolio.to_dict() for portfolio in self.portfolios.values()
        ]
        return portfolio_details, None

    def describe_portfolio_shares(
        self,
        portfolio_id: str,
        type: str,
        page_token: Optional[str] = None,
        page_size: Optional[int] = None,
    ) -> Tuple[Optional[str], List[Dict[str, Any]]]:
        """TODO: Implement pagination"""

        portfolio_share_details = []

        if portfolio_id not in self.portfolios:
            return None, []

        if type == "ACCOUNT":
            account_ids = self.portfolio_access.get(portfolio_id, [])
            for account_id in account_ids:
                portfolio_share_details.append(
                    {
                        "PrincipalId": account_id,
                        "Type": "ACCOUNT",
                        "Accepted": True,
                        "ShareTagOptions": False,
                        "SharePrincipals": False,
                    }
                )

        elif type == "ORGANIZATION":
            tokens = self.portfolio_share_tokens.get(portfolio_id, [])

            for token in tokens:
                if "ORGANIZATION" in token and "-o-" in token:
                    org_id_start = token.find("-ORGANIZATION-") + len("-ORGANIZATION-")

                    if "-tags" in token:
                        org_id_end = token.find("-tags", org_id_start)
                    elif "-principals" in token:
                        org_id_end = token.find("-principals", org_id_start)
                    else:
                        org_id_end = len(token)

                    org_id = token[org_id_start:org_id_end]

                    portfolio_share_details.append(
                        {
                            "PrincipalId": org_id,
                            "Type": "ORGANIZATION",
                            "Accepted": True,
                            "ShareTagOptions": "tags" in token,
                            "SharePrincipals": "principals" in token,
                        }
                    )

        return None, portfolio_share_details

    def describe_portfolio(
        self, accept_language: Optional[str], id: str
    ) -> Tuple[
        Dict[str, Any], List[Dict[str, str]], List[Dict[str, Any]], List[Dict[str, str]]
    ]:
        # TODO: Implement accept_language

        if id not in self.portfolios:
            return {}, [], [], []

        portfolio = self.portfolios[id]
        portfolio_detail = portfolio.to_dict()

        tags = portfolio.tags

        tag_options: List[Dict[str, Any]] = []
        budgets: List[Dict[str, Any]] = []

        return portfolio_detail, tags, tag_options, budgets


servicecatalog_backends = BackendDict(ServiceCatalogBackend, "servicecatalog")



"""
Route53 has some idiosyncrasies related to HostedZone ids
Some operations return `/hosted_zone/{id}`
Other operations return `{id}`
Some operations expect `/hosted_zone/{id}`
Other operations expect `{id}`
Some operations allow both (probably..)

The tests in this file are purely to test the different scenarios.

Note that all tests are verified against AWS.
If you want to run the tests against AWS yourself, make sure `TEST_DOMAIN_NAME` is changed to a domain name that's known to AWS.
"""

from functools import wraps
from uuid import uuid4

import boto3
import pytest
from botocore.exceptions import ClientError

from moto import mock_aws
from tests import allow_aws_request

TEST_DOMAIN_NAME = "mototests.click"
TEST_HOSTED_ZONE_NAME = f"test.{TEST_DOMAIN_NAME}"


def route53_aws_verified(func):
    @wraps(func)
    def pagination_wrapper(**kwargs):
        def create_hosted_zone():
            client = boto3.client("route53", "us-east-1")

            hosted_zone = client.create_hosted_zone(
                Name=TEST_HOSTED_ZONE_NAME, CallerReference=str(uuid4())
            )["HostedZone"]

            kwargs["hosted_zone"] = hosted_zone

            try:
                return func(**kwargs)
            finally:
                client.delete_hosted_zone(Id=hosted_zone["Id"])

        if allow_aws_request():
            return create_hosted_zone()
        else:
            with mock_aws():
                return create_hosted_zone()

    return pagination_wrapper


@pytest.mark.aws_verified
@route53_aws_verified
def test_hosted_zone_id_in_change_tags(hosted_zone=None):
    client = boto3.client("route53", "us-east-1")

    full_zone_id = hosted_zone["Id"]
    assert full_zone_id.startswith("/hostedzone/")

    # Can't use full ID here
    with pytest.raises(ClientError) as exc:
        client.change_tags_for_resource(
            ResourceType="hostedzone",
            ResourceId=full_zone_id,
            AddTags=[{"Key": "foo", "Value": "bar"}],
        )
    err = exc.value.response["Error"]
    assert err["Code"] == "InvalidInput"
    assert (
        err["Message"]
        == f"1 validation error detected: Value '{full_zone_id}' at 'resourceId' failed to satisfy constraint: Member must have length less than or equal to 32"
    )

    # if we naively limit the id-length to 32, we get a NoSuchHostedZone-exception (as expected)
    with pytest.raises(ClientError) as exc:
        client.change_tags_for_resource(
            ResourceType="hostedzone",
            ResourceId=full_zone_id[0:32],
            AddTags=[{"Key": "foo", "Value": "bar"}],
        )
    err = exc.value.response["Error"]
    assert err["Code"] == "NoSuchHostedZone"

    # Need to strip the '/hosted_zone/'-prefix
    id_without_prefix = full_zone_id.replace("/hostedzone/", "")
    client.change_tags_for_resource(
        ResourceType="hostedzone",
        ResourceId=id_without_prefix,
        AddTags=[{"Key": "foo", "Value": "bar"}],
    )

    # Test retrieval of tags with full ID
    with pytest.raises(ClientError) as exc:
        client.list_tags_for_resource(
            ResourceType="hostedzone", ResourceId=full_zone_id
        )
    err = exc.value.response["Error"]
    assert err["Code"] == "InvalidInput"
    assert (
        err["Message"]
        == f"1 validation error detected: Value '{full_zone_id}' at 'resourceId' failed to satisfy constraint: Member must have length less than or equal to 32"
    )

    # Test retrieval of tags with stripped ID
    tags = client.list_tags_for_resource(
        ResourceType="hostedzone", ResourceId=id_without_prefix
    )["ResourceTagSet"]
    assert tags == {
        "ResourceId": id_without_prefix,
        "ResourceType": "hostedzone",
        "Tags": [{"Key": "foo", "Value": "bar"}],
    }



"""Unit tests for timestreaminfluxdb-supported APIs."""

import boto3
import pytest
from botocore.exceptions import ClientError

from moto import mock_aws

# See our Development Tips on writing tests for hints on how to write good tests:
# http://docs.getmoto.org/en/latest/docs/contributing/development_tips/tests.html


"""Unit tests for timestreaminfluxdb-supported APIs."""


@mock_aws
def test_create_db_instance_success():
    client = boto3.client("timestream-influxdb", region_name="us-east-1")

    response = client.create_db_instance(
        name="test-instance",
        password="password123",
        dbInstanceType="db.influx.medium",
        vpcSubnetIds=["subnet-0123456789abcdef0"],
        vpcSecurityGroupIds=["sg-0123456789abcdef0"],
        allocatedStorage=123,
    )

    assert response["allocatedStorage"] == 123
    assert response["dbInstanceType"] == "db.influx.medium"
    assert response["name"] == "test-instance"
    assert response["vpcSecurityGroupIds"] == ["sg-0123456789abcdef0"]
    assert response["vpcSubnetIds"] == ["subnet-0123456789abcdef0"]

    assert "arn" in response
    assert "availabilityZone" in response
    assert "dbStorageType" in response
    assert "deploymentType" in response
    assert "endpoint" in response
    assert "id" in response
    assert "influxAuthParametersSecretArn" in response
    assert "networkType" in response
    assert "port" in response
    assert "publiclyAccessible" in response
    assert "secondaryAvailabilityZone" in response
    assert "status" in response
    assert "logDeliveryConfiguration" in response


@mock_aws
def test_create_db_instance_duplicate_identifier():
    client = boto3.client("timestream-influxdb", region_name="us-east-1")

    client.create_db_instance(
        name="test-instance",
        password="password123",
        dbInstanceType="db.influx.medium",
        vpcSubnetIds=["subnet-0123456789abcdef0"],
        vpcSecurityGroupIds=["sg-0123456789abcdef0"],
        allocatedStorage=123,
    )

    with pytest.raises(ClientError) as exc:
        client.create_db_instance(
            name="test-instance",
            password="password123",
            dbInstanceType="db.influx.medium",
            vpcSubnetIds=["subnet-0123456789abcdef0"],
            vpcSecurityGroupIds=["sg-0123456789abcdef0"],
            allocatedStorage=123,
        )

    assert exc.value.response["Error"]["Code"] == "ConflictException"


@mock_aws
def test_create_db_instance_name_invalid():
    invalid_names = [
        "1muststartwithletter",
        "cantendwithhyphen_",
        "nodoublehyphen--",
        "no",
        "longlonglonglonglonglonglonglonglonglongname",
    ]
    client = boto3.client("timestream-influxdb", region_name="us-east-1")

    for invalid_name in invalid_names:
        with pytest.raises(ClientError) as exc:
            client.create_db_instance(
                name="invalid_name",
                password="password123",
                dbInstanceType="db.influx.medium",
                vpcSubnetIds=["subnet-0123456789abcdef0"],
                vpcSecurityGroupIds=["sg-0123456789abcdef0"],
                allocatedStorage=123,
            )
        assert exc.value.response["Error"]["Code"] == "ValidationException"


@mock_aws
def test_create_db_instance_invalid_storage_type():
    client = boto3.client("timestream-influxdb", region_name="us-east-1")
    with pytest.raises(ClientError) as exc:
        client.create_db_instance(
            name="test-instance",
            password="password123",
            dbInstanceType="db.influx.medium",
            vpcSubnetIds=["subnet-0123456789abcdef0"],
            vpcSecurityGroupIds=["sg-0123456789abcdef0"],
            allocatedStorage=123,
            dbStorageType="invalid",
        )

        assert exc.value.response["Error"]["Code"] == "ValidationException"


@mock_aws
def test_create_db_instance_invalid_instance_type():
    client = boto3.client("timestream-influxdb", region_name="us-east-1")

    with pytest.raises(ClientError) as exc:
        client.create_db_instance(
            name="test-instance",
            password="password123",
            vpcSubnetIds=["subnet-0123456789abcdef0"],
            vpcSecurityGroupIds=["sg-0123456789abcdef0"],
            allocatedStorage=123,
            dbInstanceType="invalid",
        )

    assert exc.value.response["Error"]["Code"] == "ValidationException"


@mock_aws
def test_delete_db_instance():
    client = boto3.client("timestream-influxdb", region_name="us-east-1")

    response = client.create_db_instance(
        name="test-instance",
        password="password123",
        dbInstanceType="db.influx.medium",
        vpcSubnetIds=["subnet-0123456789abcdef0"],
        vpcSecurityGroupIds=["sg-0123456789abcdef0"],
        allocatedStorage=123,
    )

    id = response["id"]

    response = client.delete_db_instance(identifier=id)
    assert response["allocatedStorage"] == 123
    assert response["dbInstanceType"] == "db.influx.medium"
    assert response["name"] == "test-instance"
    assert response["vpcSecurityGroupIds"] == ["sg-0123456789abcdef0"]
    assert response["vpcSubnetIds"] == ["subnet-0123456789abcdef0"]

    assert "arn" in response
    assert "availabilityZone" in response
    assert "dbStorageType" in response
    assert "deploymentType" in response
    assert "endpoint" in response
    assert "id" in response
    assert "influxAuthParametersSecretArn" in response
    assert "networkType" in response
    assert "port" in response
    assert "publiclyAccessible" in response
    assert "secondaryAvailabilityZone" in response
    assert "logDeliveryConfiguration" in response

    assert response["status"] == "DELETING"


@mock_aws
def test_delete_db_instance_invalid_name():
    client = boto3.client("timestream-influxdb", region_name="us-east-1")

    id = "-100000"
    with pytest.raises(ClientError) as exc:
        client.delete_db_instance(identifier=id)

    assert exc.value.response["Error"]["Code"] == "ResourceNotFoundException"


@mock_aws
def test_get_db_instance():
    client = boto3.client("timestream-influxdb", region_name="us-east-1")

    created_instance = client.create_db_instance(
        name="test-instance",
        password="password123",
        dbInstanceType="db.influx.medium",
        vpcSubnetIds=["subnet-0123456789abcdef0"],
        vpcSecurityGroupIds=["sg-0123456789abcdef0"],
        allocatedStorage=123,
    )

    id = created_instance["id"]
    response = client.get_db_instance(identifier=id)
    assert response["allocatedStorage"] == 123
    assert response["dbInstanceType"] == "db.influx.medium"
    assert response["name"] == "test-instance"
    assert response["vpcSecurityGroupIds"] == ["sg-0123456789abcdef0"]
    assert response["vpcSubnetIds"] == ["subnet-0123456789abcdef0"]

    assert "arn" in response
    assert "availabilityZone" in response
    assert "dbStorageType" in response
    assert "deploymentType" in response
    assert "endpoint" in response
    assert "id" in response
    assert "influxAuthParametersSecretArn" in response
    assert "networkType" in response
    assert "port" in response
    assert "publiclyAccessible" in response
    assert "secondaryAvailabilityZone" in response
    assert "logDeliveryConfiguration" in response
    assert "status" in response


@mock_aws
def test_get_db_instance_invalid():
    client = boto3.client("timestream-influxdb", region_name="us-east-1")

    id = "-100000"
    with pytest.raises(ClientError) as exc:
        client.get_db_instance(identifier=id)

    assert exc.value.response["Error"]["Code"] == "ResourceNotFoundException"


@mock_aws
def test_list_db_instances_empty():
    client = boto3.client("timestream-influxdb", region_name="us-east-1")
    response = client.list_db_instances()
    assert len(response["items"]) == 0


@mock_aws
def test_list_db_instances():
    client = boto3.client("timestream-influxdb", region_name="us-east-1")
    client.create_db_instance(
        name="test-instance1",
        password="password123",
        dbInstanceType="db.influx.medium",
        vpcSubnetIds=["subnet-0123456789abcdef0"],
        vpcSecurityGroupIds=["sg-0123456789abcdef0"],
        allocatedStorage=123,
    )

    client.create_db_instance(
        name="test-instance2",
        password="password123",
        dbInstanceType="db.influx.medium",
        vpcSubnetIds=["subnet-0123456789abcdef0"],
        vpcSecurityGroupIds=["sg-0123456789abcdef0"],
        allocatedStorage=123,
    )

    response = client.list_db_instances()
    resources = response["items"]
    assert len(resources) == 2

    for resource in resources:
        assert "allocatedStorage" in resource
        assert "arn" in resource
        assert "dbInstanceType" in resource
        assert "dbStorageType" in resource
        assert "deploymentType" in resource
        assert "endpoint" in resource
        assert "id" in resource
        assert "name" in resource
        assert "networkType" in resource
        assert "port" in resource
        assert "status" in resource

    # delete one and confirm only one left
    client.delete_db_instance(identifier=resources[0]["id"])

    response = client.list_db_instances()
    resources = response["items"]
    assert len(resources) == 1


@mock_aws
def test_create_db_parameter_group():
    client = boto3.client("timestream-influxdb", region_name="us-east-1")

    response = client.create_db_parameter_group(
        name="test-parameter-group",
        description="Test parameter group for unit tests",
        parameters={
            "InfluxDBv2": {
                "fluxLogEnabled": True,
                "logLevel": "debug",
                "queryQueueSize": 100,
            }
        },
        tags={"Environment": "Test", "Project": "Moto"},
    )

    assert response["name"] == "test-parameter-group"
    assert response["description"] == "Test parameter group for unit tests"
    assert response["parameters"]["InfluxDBv2"]["fluxLogEnabled"] is True
    assert response["parameters"]["InfluxDBv2"]["logLevel"] == "debug"
    assert response["parameters"]["InfluxDBv2"]["queryQueueSize"] == 100
    assert "id" in response
    assert "arn" in response


@mock_aws
def test_get_db_parameter_group():
    client = boto3.client("timestream-influxdb", region_name="us-east-1")

    create_response = client.create_db_parameter_group(
        name="test-parameter-group",
        description="Test parameter group for unit tests",
        parameters={"InfluxDBv2": {"fluxLogEnabled": True, "logLevel": "debug"}},
    )

    param_group_id = create_response["id"]

    get_response = client.get_db_parameter_group(identifier=param_group_id)

    assert get_response["id"] == param_group_id
    assert get_response["name"] == "test-parameter-group"
    assert get_response["description"] == "Test parameter group for unit tests"
    assert get_response["parameters"]["InfluxDBv2"]["fluxLogEnabled"] is True
    assert get_response["parameters"]["InfluxDBv2"]["logLevel"] == "debug"
    assert "arn" in get_response

    with pytest.raises(ClientError) as exc:
        client.get_db_parameter_group(identifier="non-existent-id")

    assert exc.value.response["Error"]["Code"] == "ResourceNotFoundException"


@mock_aws
def test_list_db_parameter_groups():
    client = boto3.client("timestream-influxdb", region_name="us-east-1")

    list_response = client.list_db_parameter_groups()
    assert len(list_response["items"]) == 0
    assert "nextToken" not in list_response

    param_groups = []
    for i in range(5):
        response = client.create_db_parameter_group(
            name=f"test-parameter-group-{i}",
            description=f"Test parameter group {i}",
            parameters={
                "InfluxDBv2": {
                    "fluxLogEnabled": i % 2 == 0,
                    "logLevel": "debug" if i % 2 == 0 else "info",
                }
            },
        )
        param_groups.append(response)

    list_response = client.list_db_parameter_groups()
    assert len(list_response["items"]) == 5

    list_response = client.list_db_parameter_groups(maxResults=2)
    assert len(list_response["items"]) == 2
    assert "nextToken" in list_response

    next_token = list_response["nextToken"]
    list_response = client.list_db_parameter_groups(nextToken=next_token, maxResults=2)
    assert len(list_response["items"]) == 2
    assert "nextToken" in list_response

    next_token = list_response["nextToken"]
    list_response = client.list_db_parameter_groups(nextToken=next_token, maxResults=2)
    assert len(list_response["items"]) == 1
    assert "nextToken" not in list_response


@mock_aws
def test_list_db_clusters():
    client = boto3.client("timestream-influxdb", region_name="us-east-1")

    list_response = client.list_db_clusters()
    assert len(list_response["items"]) == 0
    assert "nextToken" not in list_response

    cluster_ids = []
    for i in range(5):
        response = client.create_db_cluster(
            name=f"test-cluster-{i}",
            password="password123",
            dbInstanceType="db.influx.medium",
            allocatedStorage=100,
            vpcSubnetIds=["subnet-0123456789abcdef0", "subnet-0123456789abcdef1"],
            vpcSecurityGroupIds=["sg-0123456789abcdef0"],
            deploymentType="MULTI_NODE_READ_REPLICAS",
        )
        cluster_ids.append(response["dbClusterId"])

    list_response = client.list_db_clusters()
    assert len(list_response["items"]) == 5

    for item in list_response["items"]:
        assert "id" in item
        assert "name" in item
        assert "arn" in item
        assert "status" in item
        assert "endpoint" in item
        assert "readerEndpoint" in item
        assert "port" in item
        assert "deploymentType" in item
        assert "dbInstanceType" in item
        assert "networkType" in item
        assert "dbStorageType" in item
        assert "allocatedStorage" in item

    list_response = client.list_db_clusters(maxResults=2)
    assert len(list_response["items"]) == 2
    assert "nextToken" in list_response

    next_token = list_response["nextToken"]
    list_response = client.list_db_clusters(nextToken=next_token, maxResults=2)
    assert len(list_response["items"]) == 2
    assert "nextToken" in list_response

    next_token = list_response["nextToken"]
    list_response = client.list_db_clusters(nextToken=next_token, maxResults=2)
    assert len(list_response["items"]) == 1
    assert "nextToken" not in list_response


@mock_aws
def test_get_db_cluster():
    client = boto3.client("timestream-influxdb", region_name="us-east-1")

    create_response = client.create_db_cluster(
        name="test-cluster",
        password="password123",
        dbInstanceType="db.influx.medium",
        allocatedStorage=100,
        vpcSubnetIds=["subnet-0123456789abcdef0", "subnet-0123456789abcdef1"],
        vpcSecurityGroupIds=["sg-0123456789abcdef0"],
        deploymentType="MULTI_NODE_READ_REPLICAS",
    )

    cluster_id = create_response["dbClusterId"]

    get_response = client.get_db_cluster(dbClusterId=cluster_id)

    assert get_response["id"] == cluster_id
    assert get_response["name"] == "test-cluster"
    assert get_response["status"] == "AVAILABLE"
    assert get_response["deploymentType"] == "MULTI_NODE_READ_REPLICAS"
    assert get_response["dbInstanceType"] == "db.influx.medium"
    assert get_response["allocatedStorage"] == 100
    assert isinstance(get_response["arn"], str)
    assert isinstance(get_response["endpoint"], str)
    assert isinstance(get_response["readerEndpoint"], str)
    assert isinstance(get_response["port"], int)
    assert isinstance(get_response["vpcSubnetIds"], list)
    assert isinstance(get_response["vpcSecurityGroupIds"], list)
    assert len(get_response["vpcSubnetIds"]) == 2
    assert len(get_response["vpcSecurityGroupIds"]) == 1

    with pytest.raises(ClientError) as exc:
        client.get_db_cluster(dbClusterId="non-existent-id")

    assert exc.value.response["Error"]["Code"] == "ResourceNotFoundException"


@mock_aws
def test_create_db_cluster():
    client = boto3.client("timestream-influxdb", region_name="us-east-1")

    response = client.create_db_cluster(
        name="test-cluster",
        password="password123",
        dbInstanceType="db.influx.medium",
        allocatedStorage=100,
        vpcSubnetIds=["subnet-0123456789abcdef0", "subnet-0123456789abcdef1"],
        vpcSecurityGroupIds=["sg-0123456789abcdef0"],
        deploymentType="MULTI_NODE_READ_REPLICAS",
    )

    assert "dbClusterId" in response
    assert response["dbClusterStatus"] == "AVAILABLE"

    with pytest.raises(ClientError) as exc:
        client.create_db_cluster(
            name="test-cluster",
            password="password123",
            dbInstanceType="db.influx.medium",
            allocatedStorage=100,
            vpcSubnetIds=["subnet-0123456789abcdef0", "subnet-0123456789abcdef1"],
            vpcSecurityGroupIds=["sg-0123456789abcdef0"],
            deploymentType="MULTI_NODE_READ_REPLICAS",
        )

    assert exc.value.response["Error"]["Code"] == "ConflictException"

    response = client.create_db_cluster(
        name="test-cluster-full",
        username="admin",
        password="password123",
        organization="test-org",
        bucket="test-bucket",
        port=8088,
        dbParameterGroupIdentifier="test-param-group",
        dbInstanceType="db.influx.medium",
        dbStorageType="InfluxIOIncludedT1",
        allocatedStorage=100,
        networkType="IPV4",
        publiclyAccessible=True,
        vpcSubnetIds=["subnet-0123456789abcdef0", "subnet-0123456789abcdef1"],
        vpcSecurityGroupIds=["sg-0123456789abcdef0"],
        deploymentType="MULTI_NODE_READ_REPLICAS",
        failoverMode="AUTOMATIC",
        logDeliveryConfiguration={
            "s3Configuration": {"bucketName": "test-bucket", "enabled": True}
        },
        tags={"Environment": "Test", "Project": "Moto"},
    )

    assert "dbClusterId" in response
    assert response["dbClusterStatus"] == "AVAILABLE"



from botocore.model import ServiceModel

from moto.core.responses import ActionContext, ActionResult, BaseResponse, EmptyResult
from moto.core.serialize import QuerySerializer, never_return

TEST_MODEL = {
    "metadata": {"protocol": "query", "apiVersion": "2014-01-01"},
    "documentation": "",
    "operations": {
        "TestOperation": {
            "name": "TestOperation",
            "http": {
                "method": "POST",
                "requestUri": "/",
            },
            "output": {"shape": "OutputShape"},
        }
    },
    "shapes": {
        "OutputShape": {
            "type": "structure",
            "members": {
                "string": {"shape": "StringType"},
                "TransformerTest": {"shape": "StringType"},
            },
        },
        "StringType": {
            "type": "string",
        },
    },
}


def test_response_result_execution() -> None:
    """Exercise various response result execution paths."""

    class TestResponseClass(BaseResponse):
        RESPONSE_KEY_PATH_TO_TRANSFORMER = {"OutputShape.TransformerTest": never_return}

    class TestResponseObject:
        string = "test-string"
        TransformerTest = "some data"

    service_model = ServiceModel(TEST_MODEL)
    operation_model = service_model.operation_model("TestOperation")
    serializer_class = QuerySerializer
    response_class = TestResponseClass

    context = ActionContext(
        service_model, operation_model, serializer_class, response_class
    )

    result = ActionResult(TestResponseObject())
    _, __, body = result.execute_result(context)
    assert "test-string" in body
    assert "TransformerTest" not in body

    result = EmptyResult()
    _, __, body = result.execute_result(context)
    assert "<OutputShapeResult/>" in body



import boto3
import pytest
from botocore.exceptions import ClientError

from moto import mock_aws

from . import DEFAULT_REGION
from .test_rds import (
    create_db_instance,
    create_subnet,
)
from .test_rds_clusters import create_db_cluster


@pytest.fixture(name="client")
@mock_aws
def get_rds_client():
    return boto3.client("rds", region_name=DEFAULT_REGION)


@mock_aws
def test_create_bluegreen_deployment_creates_a_green_db_instance(client):
    expected_target_engine_version = "17.3"
    expected_target_db_instance_class = "db.6g.xlarge"
    expected_target_allocated_storage = 999
    expected_db_parameter_group_name = "new_parameter_group"
    expected_target_iops = 5000
    expected_target_storage_type = "gp3"
    expected_target_storage_throughput = 6329

    subnet_id = create_subnet()
    subnet = client.create_db_subnet_group(
        DBSubnetGroupName="db_subnet1",
        DBSubnetGroupDescription="my db subnet",
        SubnetIds=[subnet_id],
        Tags=[{"Value": "bar", "Key": "foo"}, {"Value": "bar1", "Key": "foo1"}],
    )["DBSubnetGroup"]["DBSubnetGroupName"]

    instance = create_db_instance(
        DBInstanceIdentifier="FooBar",
        DBSubnetGroupName=subnet,
        MasterUsername="Bob",
        ManageMasterUserPassword=False,
    )
    client.create_db_parameter_group(
        DBParameterGroupName=expected_db_parameter_group_name,
        DBParameterGroupFamily="postgres",
        Description="Foobar",
    )

    response = client.create_blue_green_deployment(
        BlueGreenDeploymentName="FooBarBlueGreen",
        Source=instance["DBInstanceArn"],
        TargetEngineVersion=expected_target_engine_version,
        TargetDBInstanceClass=expected_target_db_instance_class,
        TargetAllocatedStorage=expected_target_allocated_storage,
        TargetDBParameterGroupName=expected_db_parameter_group_name,
        TargetIops=expected_target_iops,
        TargetStorageType=expected_target_storage_type,
        TargetStorageThroughput=expected_target_storage_throughput,
    )

    bluegreen_deployment_response = response["BlueGreenDeployment"]

    db_describe_response = client.describe_db_instances(
        DBInstanceIdentifier=bluegreen_deployment_response["Target"]
    )

    assert bluegreen_deployment_response is not None
    assert "bgd-" in bluegreen_deployment_response["BlueGreenDeploymentIdentifier"]
    assert bluegreen_deployment_response["BlueGreenDeploymentName"] == "FooBarBlueGreen"
    assert bluegreen_deployment_response["Source"] == instance["DBInstanceArn"]
    assert (
        bluegreen_deployment_response["Target"]
        == db_describe_response["DBInstances"][0]["DBInstanceArn"]
    )
    assert bluegreen_deployment_response["Status"] == "PROVISIONING"

    assert len(db_describe_response["DBInstances"]) > 0
    assert (
        db_describe_response["DBInstances"][0]["EngineVersion"]
        == expected_target_engine_version
    )
    assert (
        db_describe_response["DBInstances"][0]["DBInstanceClass"]
        == expected_target_db_instance_class
    )
    assert (
        db_describe_response["DBInstances"][0]["AllocatedStorage"]
        == expected_target_allocated_storage
    )
    assert (
        db_describe_response["DBInstances"][0]["DBParameterGroups"][0][
            "DBParameterGroupName"
        ]
        == expected_db_parameter_group_name
    )
    assert db_describe_response["DBInstances"][0]["Iops"] == expected_target_iops
    assert (
        db_describe_response["DBInstances"][0]["StorageType"]
        == expected_target_storage_type
    )
    assert (
        db_describe_response["DBInstances"][0]["StorageThroughput"]
        == expected_target_storage_throughput
    )
    assert (
        db_describe_response["DBInstances"][0]["DBSecurityGroups"]
        == instance["DBSecurityGroups"]
    )
    assert (
        db_describe_response["DBInstances"][0]["DBSubnetGroup"]["DBSubnetGroupName"]
        == instance["DBSubnetGroup"]["DBSubnetGroupName"]
    )
    assert (
        db_describe_response["DBInstances"][0]["OptionGroupMemberships"][0][
            "OptionGroupName"
        ]
        != instance["OptionGroupMemberships"][0]["OptionGroupName"]
    )
    assert db_describe_response["DBInstances"][0].get("MasterUserSecret") is None


@mock_aws
def test_create_bluegreen_deployment_creates_a_green_db_cluster(client):
    expected_target_engine_version = "17.3"
    expected_db_parameter_group_name = "new_parameter_group"
    expected_db_cluster_parameter_group_name = "new_cluster_parameter_group"

    subnet_id = create_subnet()
    subnet = client.create_db_subnet_group(
        DBSubnetGroupName="db_subnet1",
        DBSubnetGroupDescription="my db subnet",
        SubnetIds=[subnet_id],
        Tags=[{"Value": "bar", "Key": "foo"}, {"Value": "bar1", "Key": "foo1"}],
    )["DBSubnetGroup"]["DBSubnetGroupName"]

    create_db_cluster(
        DBClusterIdentifier="cluster-1",
        DBSubnetGroupName=subnet,
        MasterUsername="Bob",
        Engine="aurora-postgresql",
    )
    clustered_instances = []
    for i in range(3):
        clustered_instances.append(
            client.create_db_instance(
                DBInstanceIdentifier=f"test-instance-{i}",
                DBInstanceClass="db.m1.small",
                Engine="aurora-postgresql",
                DBClusterIdentifier="cluster-1",
                DBSubnetGroupName=subnet,
            )
        )
    cluster = client.describe_db_clusters(
        DBClusterIdentifier="cluster-1",
    )["DBClusters"][0]
    client.create_db_parameter_group(
        DBParameterGroupName=expected_db_parameter_group_name,
        DBParameterGroupFamily="postgres",
        Description="Foobar",
    )
    client.create_db_cluster_parameter_group(
        DBClusterParameterGroupName=expected_db_cluster_parameter_group_name,
        DBParameterGroupFamily="aurora-postgresql14",
        Description="FooBar",
    )

    response = client.create_blue_green_deployment(
        BlueGreenDeploymentName="FooBarBlueGreen",
        Source=cluster["DBClusterArn"],
        TargetEngineVersion=expected_target_engine_version,
        TargetDBParameterGroupName=expected_db_parameter_group_name,
        TargetDBClusterParameterGroupName=expected_db_cluster_parameter_group_name,
    )

    bluegreen_deployment_response = response["BlueGreenDeployment"]

    db_describe_response = client.describe_db_clusters(
        DBClusterIdentifier=bluegreen_deployment_response["Target"]
    )

    assert bluegreen_deployment_response is not None
    assert "bgd-" in bluegreen_deployment_response["BlueGreenDeploymentIdentifier"]
    assert bluegreen_deployment_response["BlueGreenDeploymentName"] == "FooBarBlueGreen"
    assert bluegreen_deployment_response["Source"] == cluster["DBClusterArn"]
    assert (
        bluegreen_deployment_response["Target"]
        == db_describe_response["DBClusters"][0]["DBClusterArn"]
    )
    assert bluegreen_deployment_response["Status"] == "PROVISIONING"

    assert len(db_describe_response["DBClusters"]) > 0
    assert (
        db_describe_response["DBClusters"][0]["EngineVersion"]
        == expected_target_engine_version
    )
    assert (
        db_describe_response["DBClusters"][0]["AllocatedStorage"]
        == cluster["AllocatedStorage"]
    )
    assert db_describe_response["DBClusters"][0].get("MasterUserSecret") is None
    assert (
        db_describe_response["DBClusters"][0]["DBClusterParameterGroup"]
        == expected_db_cluster_parameter_group_name
    )

    assert len(db_describe_response["DBClusters"][0]["DBClusterMembers"]) == len(
        cluster["DBClusterMembers"]
    )
    for index, member in enumerate(
        db_describe_response["DBClusters"][0]["DBClusterMembers"]
    ):
        member_response = client.describe_db_instances(
            DBInstanceIdentifier=member["DBInstanceIdentifier"]
        )

        assert (
            member_response["DBInstances"][0]["DBSubnetGroup"]["DBSubnetGroupName"]
            == clustered_instances[index]["DBInstance"]["DBSubnetGroup"][
                "DBSubnetGroupName"
            ]
        )


@mock_aws
def test_create_blue_green_deployment_db_instance_with_managed_master_password(client):
    instance = create_db_instance(
        DBInstanceIdentifier="FooBar",
        MasterUsername="Bob",
        ManageMasterUserPassword=True,
    )

    source_arn = instance["DBInstanceArn"]

    with pytest.raises(ClientError) as exc:
        client.create_blue_green_deployment(
            BlueGreenDeploymentName="FooBarBlueGreen",
            Source=source_arn,
        )

    err = exc.value.response["Error"]
    assert err["Code"] == "SourceDatabaseNotSupportedFault"
    assert (
        err["Message"]
        == f"The source DB instance {source_arn} isn't supported for a blue/green deployment."
    )


@mock_aws
def test_create_blue_green_deployment_rds_cluster_with_managed_master_password(client):
    create_db_cluster(
        DBClusterIdentifier="cluster-1",
        MasterUsername="Bob",
        ManageMasterUserPassword=True,
        Engine="aurora-postgresql",
    )
    clustered_instances = []
    for i in range(3):
        clustered_instances.append(
            client.create_db_instance(
                DBInstanceIdentifier=f"test-instance-{i}",
                DBInstanceClass="db.m1.small",
                Engine="aurora-postgresql",
                DBClusterIdentifier="cluster-1",
            )
        )
    cluster = client.describe_db_clusters(
        DBClusterIdentifier="cluster-1",
    )["DBClusters"][0]

    source_arn = cluster["DBClusterArn"]

    with pytest.raises(ClientError) as exc:
        client.create_blue_green_deployment(
            BlueGreenDeploymentName="FooBarBlueGreen",
            Source=source_arn,
        )

    err = exc.value.response["Error"]
    assert err["Code"] == "SourceDatabaseNotSupportedFault"
    assert (
        err["Message"]
        == f"The source DB instance {source_arn} isn't supported for a blue/green deployment."
    )


@mock_aws
def test_create_blue_green_deployment_error_if_source_db_instance_not_found(client):
    with pytest.raises(ClientError) as ex:
        client.create_blue_green_deployment(
            BlueGreenDeploymentName="FooBarBlueGreen",
            Source="arn:rds:123456789012:us-east-1:db:not_an_arn",
        )

    err = ex.value.response["Error"]
    assert err["Code"] == "DBInstanceNotFound"
    assert (
        err["Message"]
        == "DBInstance arn:rds:123456789012:us-east-1:db:not_an_arn not found."
    )


@mock_aws
def test_create_blue_green_deployment_error_if_source_rds_cluster_not_found(client):
    with pytest.raises(ClientError) as ex:
        client.create_blue_green_deployment(
            BlueGreenDeploymentName="FooBarBlueGreen",
            Source="arn:rds:123456789012:us-east-1:cluster:not_an_arn",
        )

    err = ex.value.response["Error"]
    assert err["Code"] == "DBClusterNotFoundFault"
    assert (
        err["Message"]
        == "DBCluster arn:rds:123456789012:us-east-1:cluster:not_an_arn not found."
    )


@mock_aws
def test_create_blue_green_deployment_with_duplicate_name(client):
    bluegreen_name = "FooBarBlueGreen"
    instance = create_db_instance(
        DBInstanceIdentifier="FooBar",
    )

    response = client.create_blue_green_deployment(
        BlueGreenDeploymentName=bluegreen_name,
        Source=instance["DBInstanceArn"],
    )

    with pytest.raises(ClientError) as ex:
        client.create_blue_green_deployment(
            BlueGreenDeploymentName=bluegreen_name,
            Source=instance["DBInstanceArn"],
        )

    error = ex.value.response["Error"]

    assert response["BlueGreenDeployment"]["BlueGreenDeploymentName"] == bluegreen_name
    assert error["Code"] == "BlueGreenDeploymentAlreadyExistsFault"
    assert (
        error["Message"]
        == "A blue/green deployment with the specified name FooBarBlueGreen already exists."
    )


@mock_aws
def test_describe_bluegreen_deployments_db_instance_with_filters(client):
    bg_name = "bluegreen-deployment-1"
    instance = create_db_instance(
        DBInstanceIdentifier="FooBar",
    )

    create_response = client.create_blue_green_deployment(
        BlueGreenDeploymentName=bg_name,
        Source=instance["DBInstanceArn"],
    )

    describe_response = client.describe_blue_green_deployments(
        BlueGreenDeploymentIdentifier=create_response["BlueGreenDeployment"][
            "BlueGreenDeploymentIdentifier"
        ]
    )

    describe_with_filters_response = client.describe_blue_green_deployments(
        Filters=[
            {"Name": "blue-green-deployment-name", "Values": [bg_name]},
            {"Name": "source", "Values": [instance["DBInstanceArn"]]},
        ]
    )

    update_status_from_create_blue_green_response(
        create_response["BlueGreenDeployment"]
    )

    assert (
        describe_response["BlueGreenDeployments"][0]["Tasks"][0]["Status"]
        == "COMPLETED"
    )

    assert (
        describe_response["BlueGreenDeployments"][0]
        == create_response["BlueGreenDeployment"]
    )
    assert len(describe_with_filters_response["BlueGreenDeployments"]) > 0
    assert (
        describe_with_filters_response["BlueGreenDeployments"][0]
        == describe_response["BlueGreenDeployments"][0]
    )


@mock_aws
def test_describe_bluegreen_deployments_db_cluster_with_filters(client):
    bg_name = "bluegreen-deployment-1"
    create_db_cluster(
        DBClusterIdentifier="cluster-1",
    )

    cluster = client.describe_db_clusters(
        DBClusterIdentifier="cluster-1",
    )["DBClusters"][0]

    create_response = client.create_blue_green_deployment(
        BlueGreenDeploymentName=bg_name,
        Source=cluster["DBClusterArn"],
    )

    describe_response = client.describe_blue_green_deployments(
        BlueGreenDeploymentIdentifier=create_response["BlueGreenDeployment"][
            "BlueGreenDeploymentIdentifier"
        ]
    )

    describe_with_filters_response = client.describe_blue_green_deployments(
        Filters=[
            {"Name": "blue-green-deployment-name", "Values": [bg_name]},
            {"Name": "source", "Values": [cluster["DBClusterArn"]]},
        ]
    )

    update_status_from_create_blue_green_response(
        create_response["BlueGreenDeployment"]
    )

    assert (
        describe_response["BlueGreenDeployments"][0]["Tasks"][0]["Status"]
        == "COMPLETED"
    )

    assert (
        describe_response["BlueGreenDeployments"][0]
        == create_response["BlueGreenDeployment"]
    )
    assert len(describe_with_filters_response["BlueGreenDeployments"]) > 0
    assert (
        describe_with_filters_response["BlueGreenDeployments"][0]
        == create_response["BlueGreenDeployment"]
    )


@mock_aws
def test_describe_bluegreen_deployments_with_no_hits(client):
    with pytest.raises(ClientError) as ex:
        client.describe_blue_green_deployments(
            BlueGreenDeploymentIdentifier="red-yellow"
        )

    filter_response = client.describe_blue_green_deployments(
        Filters=[{"Name": "source", "Values": ["no-real-arn"]}]
    )

    error = ex.value.response["Error"]

    assert error["Code"] == "BlueGreenDeploymentNotFoundFault"
    assert len(filter_response["BlueGreenDeployments"]) == 0


@mock_aws
def test_switchover_blue_green_deployment_switches_db_instances(client):
    bg_name = "bluegreen-deployment-1"
    instance = create_db_instance(
        DBInstanceIdentifier="FooBar",
    )

    create_response = client.create_blue_green_deployment(
        BlueGreenDeploymentName=bg_name,
        Source=instance["DBInstanceArn"],
    )

    switchover_response = client.switchover_blue_green_deployment(
        BlueGreenDeploymentIdentifier=create_response["BlueGreenDeployment"][
            "BlueGreenDeploymentIdentifier"
        ]
    )

    describe_response = client.describe_blue_green_deployments(
        BlueGreenDeploymentIdentifier=create_response["BlueGreenDeployment"][
            "BlueGreenDeploymentIdentifier"
        ]
    )

    describe_target_instance = client.describe_db_instances(
        DBInstanceIdentifier=describe_response["BlueGreenDeployments"][0]["Target"]
    )

    assert (
        switchover_response["BlueGreenDeployment"]["Status"] == "SWITCHOVER_IN_PROGRESS"
    )
    assert (
        switchover_response["BlueGreenDeployment"]["Source"]
        == instance["DBInstanceArn"]
    )
    assert "green" in switchover_response["BlueGreenDeployment"]["Target"]

    assert (
        describe_response["BlueGreenDeployments"][0]["Status"] == "SWITCHOVER_COMPLETED"
    )
    assert "old" in describe_response["BlueGreenDeployments"][0]["Source"]
    assert (
        describe_response["BlueGreenDeployments"][0]["Target"].split(":")[-1]
        == instance["DBInstanceIdentifier"]
    )
    assert (
        describe_target_instance["DBInstances"][0]["Endpoint"] == instance["Endpoint"]
    )


@mock_aws
def test_switchover_blue_green_deployment_switches_rds_clusters(client):
    bg_name = "bluegreen-deployment-1"
    create_db_cluster(
        DBClusterIdentifier="cluster-1",
    )
    cluster = client.describe_db_clusters(
        DBClusterIdentifier="cluster-1",
    )["DBClusters"][0]

    create_response = client.create_blue_green_deployment(
        BlueGreenDeploymentName=bg_name,
        Source=cluster["DBClusterArn"],
    )

    switchover_response = client.switchover_blue_green_deployment(
        BlueGreenDeploymentIdentifier=create_response["BlueGreenDeployment"][
            "BlueGreenDeploymentIdentifier"
        ]
    )

    describe_response = client.describe_blue_green_deployments(
        BlueGreenDeploymentIdentifier=create_response["BlueGreenDeployment"][
            "BlueGreenDeploymentIdentifier"
        ]
    )

    describe_target_instance = client.describe_db_clusters(
        DBClusterIdentifier=describe_response["BlueGreenDeployments"][0]["Target"]
    )

    assert (
        switchover_response["BlueGreenDeployment"]["Status"] == "SWITCHOVER_IN_PROGRESS"
    )
    assert (
        switchover_response["BlueGreenDeployment"]["Source"] == cluster["DBClusterArn"]
    )
    assert "green" in switchover_response["BlueGreenDeployment"]["Target"]

    assert (
        describe_response["BlueGreenDeployments"][0]["Status"] == "SWITCHOVER_COMPLETED"
    )
    assert "old" in describe_response["BlueGreenDeployments"][0]["Source"]
    assert (
        describe_response["BlueGreenDeployments"][0]["Target"].split(":")[-1]
        == cluster["DBClusterIdentifier"]
    )
    assert (
        describe_target_instance["DBClusters"][0]["Endpoint"].split(".")[0]
        == cluster["Endpoint"].split(".")[0]
    )


@mock_aws
def test_switchover_after_successful_switchover_blue_green_deployment(client):
    bg_name = "bluegreen-deployment-1"
    instance = create_db_instance(
        DBInstanceIdentifier="FooBar",
    )

    create_response = client.create_blue_green_deployment(
        BlueGreenDeploymentName=bg_name,
        Source=instance["DBInstanceArn"],
    )

    client.switchover_blue_green_deployment(
        BlueGreenDeploymentIdentifier=create_response["BlueGreenDeployment"][
            "BlueGreenDeploymentIdentifier"
        ]
    )

    with pytest.raises(ClientError) as exc:
        client.switchover_blue_green_deployment(
            BlueGreenDeploymentIdentifier=create_response["BlueGreenDeployment"][
                "BlueGreenDeploymentIdentifier"
            ]
        )

    error = exc.value.response["Error"]
    assert error["Code"] == "InvalidBlueGreenDeploymentStateFault"


@mock_aws
@pytest.mark.parametrize(
    "options",
    [
        pytest.param(
            {"with_switchover": False, "delete_target": True},
            id="before_switchover_with_delete_target",
        ),
        pytest.param(
            {"with_switchover": False, "delete_target": False},
            id="before_switchover_without_delete_target",
        ),
        pytest.param(
            {"with_switchover": True, "delete_target": True},
            id="after_switchover_with_delete_target",
        ),
    ],
)
def test_delete_blue_green_deployment_with_source_db_instance(client, options):
    bg_name = "bluegreen-deployment-1"

    instance = create_db_instance(
        DBInstanceIdentifier="FooBar",
    )

    create_response = client.create_blue_green_deployment(
        BlueGreenDeploymentName=bg_name,
        Source=instance["DBInstanceArn"],
    )

    if options["with_switchover"]:
        client.switchover_blue_green_deployment(
            BlueGreenDeploymentIdentifier=create_response["BlueGreenDeployment"][
                "BlueGreenDeploymentIdentifier"
            ]
        )

    delete_response = client.delete_blue_green_deployment(
        BlueGreenDeploymentIdentifier=create_response["BlueGreenDeployment"][
            "BlueGreenDeploymentIdentifier"
        ],
        DeleteTarget=options["delete_target"],
    )

    describe_response = client.describe_blue_green_deployments(
        Filters=[{"Name": "blue-green-deployment-name", "Values": [bg_name]}]
    )

    describe_source_instance = client.describe_db_instances(
        DBInstanceIdentifier=delete_response["BlueGreenDeployment"]["Source"]
    )

    describe_target_instance = client.describe_db_instances(
        Filters=[
            {
                "Name": "db-instance-id",
                "Values": [delete_response["BlueGreenDeployment"]["Target"]],
            }
        ]
    )

    assert delete_response["BlueGreenDeployment"]["Status"] == "DELETING"
    assert len(describe_response["BlueGreenDeployments"]) == 0
    assert len(describe_source_instance["DBInstances"]) == 1
    if options["delete_target"] and not options["with_switchover"]:
        assert len(describe_target_instance["DBInstances"]) == 0
    else:
        assert len(describe_target_instance["DBInstances"]) == 1


@mock_aws
def test_delete_blue_green_deployment_when_it_does_not_exist(client):
    with pytest.raises(ClientError) as exc:
        client.delete_blue_green_deployment(
            BlueGreenDeploymentIdentifier="DoesNotExist"
        )

    error = exc.value.response["Error"]
    assert error["Code"] == "BlueGreenDeploymentNotFoundFault"


@mock_aws
@pytest.mark.parametrize(
    "options",
    [
        pytest.param(
            {"with_switchover": False, "delete_target": True},
            id="before_switchover_with_delete_target",
        ),
        pytest.param(
            {"with_switchover": False, "delete_target": False},
            id="before_switchover_without_delete_target",
        ),
        pytest.param(
            {"with_switchover": True, "delete_target": True},
            id="after_switchover_with_delete_target",
        ),
    ],
)
def test_delete_blue_green_deployment_with_source_db_cluster(client, options):
    bg_name = "bluegreen-deployment-1"

    create_db_cluster(
        DBClusterIdentifier="cluster-1",
        Engine="aurora-postgresql",
    )
    clustered_instances = []
    for i in range(3):
        clustered_instances.append(
            client.create_db_instance(
                DBInstanceIdentifier=f"test-instance-{i}",
                DBClusterIdentifier="cluster-1",
                DBInstanceClass="db.m1.small",
                Engine="aurora-postgresql",
            )
        )
    cluster = client.describe_db_clusters(
        DBClusterIdentifier="cluster-1",
    )["DBClusters"][0]

    create_response = client.create_blue_green_deployment(
        BlueGreenDeploymentName=bg_name,
        Source=cluster["DBClusterArn"],
    )

    if options["with_switchover"]:
        client.switchover_blue_green_deployment(
            BlueGreenDeploymentIdentifier=create_response["BlueGreenDeployment"][
                "BlueGreenDeploymentIdentifier"
            ]
        )

    delete_response = client.delete_blue_green_deployment(
        BlueGreenDeploymentIdentifier=create_response["BlueGreenDeployment"][
            "BlueGreenDeploymentIdentifier"
        ],
        DeleteTarget=options["delete_target"],
    )

    describe_response = client.describe_blue_green_deployments(
        Filters=[{"Name": "blue-green-deployment-name", "Values": [bg_name]}]
    )

    describe_source_cluster = client.describe_db_clusters(
        DBClusterIdentifier=delete_response["BlueGreenDeployment"]["Source"]
    )

    describe_target_cluster = client.describe_db_clusters(
        Filters=[
            {
                "Name": "db-cluster-id",
                "Values": [delete_response["BlueGreenDeployment"]["Target"]],
            }
        ]
    )

    assert delete_response["BlueGreenDeployment"]["Status"] == "DELETING"
    assert len(describe_response["BlueGreenDeployments"]) == 0
    assert len(describe_source_cluster["DBClusters"]) == 1
    if options["delete_target"] and not options["with_switchover"]:
        assert len(describe_target_cluster["DBClusters"]) == 0
    else:
        assert len(describe_target_cluster["DBClusters"]) == 1


def update_status_from_create_blue_green_response(bg_response: dict) -> None:
    bg_response["Status"] = "AVAILABLE"

    for details in bg_response["SwitchoverDetails"]:
        details["Status"] = "AVAILABLE"
    for task in bg_response["Tasks"]:
        task["Status"] = "COMPLETED"



import json
import os
from functools import wraps
from unittest import SkipTest
from uuid import uuid4

import boto3
import botocore

from moto import mock_aws
from tests import allow_aws_request


def sns_aws_verified(func):
    """
    Function that is verified to work against AWS.
    Can be run against AWS at any time by setting:
      MOTO_TEST_ALLOW_AWS_REQUEST=true

    If this environment variable is not set, the function runs in a `mock_aws` context.
    """

    @wraps(func)
    def pagination_wrapper():
        allow_aws_request = (
            os.environ.get("MOTO_TEST_ALLOW_AWS_REQUEST", "false").lower() == "true"
        )

        if allow_aws_request:
            ssm = boto3.client("ssm", "us-east-1")
            try:
                param = ssm.get_parameter(
                    Name="/moto/tests/ses/firebase_api_key", WithDecryption=True
                )
                api_key = param["Parameter"]["Value"]
                resp = func(api_key)
            except botocore.exceptions.ClientError:
                # SNS tests try to create a PlatformApplication that connects to GCM
                # (Google Cloud Messaging, also known as Firebase Messaging)
                # That requires an account and an API key
                # If the API key has not been configured in SSM, we'll just skip the test
                #
                # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/sns/client/create_platform_application.html
                # AWS calls it 'API key', but Firebase calls it Server Key
                #
                # https://stackoverflow.com/a/75896532/13245310
                raise SkipTest("Can't execute SNS tests without Firebase API key")
        else:
            with mock_aws():
                resp = func("mock_api_key")
        return resp

    return pagination_wrapper


def sns_sqs_aws_verified(fifo_topic: bool = False, fifo_queue: bool = False):
    """
    Function that is verified to work against AWS.
    Can be run against AWS at any time by setting:
      MOTO_TEST_ALLOW_AWS_REQUEST=true

    If this environment variable is not set, the function runs in a `mock_aws` context.

    This decorator will:
      - Create an SNS topic + SQS queue
      - Run the test, passing both names as an argument
      - Delete both resources
    """

    def inner(func):
        @wraps(func)
        def pagination_wrapper(**kwargs):
            topic_name = f"test_topic_{str(uuid4())[0:6]}" + (
                ".fifo" if fifo_topic else ""
            )
            kwargs["topic_name"] = topic_name
            queue_name = f"test_queue_{str(uuid4())[0:6]}" + (
                ".fifo" if fifo_queue else ""
            )
            kwargs["queue_name"] = queue_name

            def create_resources_and_test():
                client = boto3.client("sns", region_name="us-east-1")
                sqs_client = boto3.resource("sqs", region_name="us-east-1")

                topic_args = {}
                if fifo_topic:
                    topic_args["Attributes"] = {
                        "FifoTopic": "true",
                        "ContentBasedDeduplication": "true",
                    }
                topic_arn = client.create_topic(Name=topic_name, **topic_args)[
                    "TopicArn"
                ]
                kwargs["topic_arn"] = topic_arn

                queue_args = {}
                if fifo_queue:
                    queue_args["Attributes"] = {
                        "FifoQueue": "true",
                        "ContentBasedDeduplication": "true",
                    }
                queue = sqs_client.create_queue(QueueName=queue_name, **queue_args)

                identity = boto3.client("sts", "us-east-1").get_caller_identity()
                kwargs["queue_arn"] = (
                    f"arn:aws:sqs:us-east-1:{identity['Account']}:{queue_name}"
                )

                # Give permissions to the queue
                policy = {
                    "Version": "2012-10-17",
                    "Statement": [
                        {
                            "Effect": "Allow",
                            "Principal": {"Service": "sns.amazonaws.com"},
                            "Action": "sqs:SendMessage",
                            "Resource": kwargs["queue_arn"],
                            "Condition": {"ArnEquals": {"aws:SourceArn": topic_arn}},
                        }
                    ],
                }
                queue.set_attributes(Attributes={"Policy": json.dumps(policy)})

                subscription_arn = None
                try:
                    subscription_arn = client.subscribe(
                        TopicArn=topic_arn,
                        Protocol="sqs",
                        Endpoint=kwargs["queue_arn"],
                    )["SubscriptionArn"]

                    resp = func(**kwargs)
                finally:
                    if subscription_arn:
                        client.unsubscribe(SubscriptionArn=subscription_arn)
                    client.delete_topic(TopicArn=topic_arn)
                    queue.delete()

                return resp

            if allow_aws_request():
                return create_resources_and_test()
            else:
                with mock_aws():
                    return create_resources_and_test()

        return pagination_wrapper

    return inner



"""SecurityHubBackend class with methods for supported APIs."""

import datetime
from typing import Any, Dict, List, Optional, Tuple

from moto.core.base_backend import BackendDict, BaseBackend
from moto.core.common_models import BaseModel
from moto.core.exceptions import RESTError
from moto.organizations.exceptions import AWSOrganizationsNotInUseException
from moto.organizations.models import organizations_backends
from moto.securityhub.exceptions import InvalidInputException
from moto.utilities.paginator import paginate


class Finding(BaseModel):
    def __init__(self, finding_id: str, finding_data: Dict[str, Any]):
        self.id = finding_id
        self.data = finding_data

    def as_dict(self) -> Dict[str, Any]:
        return self.data


class SecurityHubBackend(BaseBackend):
    """Implementation of SecurityHub APIs."""

    PAGINATION_MODEL = {
        "get_findings": {
            "input_token": "next_token",
            "limit_key": "max_results",
            "limit_default": 100,
            "unique_attribute": "Id",
            "fail_on_invalid_token": True,
        }
    }

    _org_configs: Dict[str, Dict[str, Any]] = {}

    def __init__(self, region_name: str, account_id: str):
        super().__init__(region_name, account_id)
        self.findings: List[Finding] = []
        self.region_name = region_name
        self.org_backend = organizations_backends[self.account_id]["aws"]

    def _get_org_config(self) -> Dict[str, Any]:
        """Get organization config for the current account."""
        try:
            org = self.org_backend.describe_organization()
            org_id = org["Organization"]["Id"]
        except RESTError:
            raise AWSOrganizationsNotInUseException()

        if org_id not in SecurityHubBackend._org_configs:
            SecurityHubBackend._org_configs[org_id] = {
                "admin_account_id": None,
                "auto_enable": False,
                "auto_enable_standards": "DEFAULT",
                "configuration": {
                    "ConfigurationType": "LOCAL",
                    "Status": "ENABLED",
                    "StatusMessage": "",
                },
            }
        return SecurityHubBackend._org_configs[org_id]

    @paginate(pagination_model=PAGINATION_MODEL)
    def get_findings(
        self,
        filters: Optional[Dict[str, Any]] = None,
        sort_criteria: Optional[List[Dict[str, str]]] = None,
        max_results: Optional[int] = None,
        next_token: Optional[str] = None,
    ) -> List[Dict[str, str]]:
        """
        Returns findings based on optional filters and sort criteria.
        """
        if max_results is not None:
            try:
                max_results = int(max_results)
                if max_results < 1 or max_results > 100:
                    raise InvalidInputException(
                        op="GetFindings",
                        msg="MaxResults must be a number between 1 and 100",
                    )
            except ValueError:
                raise InvalidInputException(
                    op="GetFindings", msg="MaxResults must be a number greater than 0"
                )

        findings = self.findings

        # TODO: Apply filters if provided
        # TODO: Apply sort criteria if provided

        return [f.as_dict() for f in findings]

    def batch_import_findings(
        self, findings: List[Dict[str, Any]]
    ) -> Tuple[int, int, List[Dict[str, Any]]]:
        """
        Import findings in batch to SecurityHub.

        Args:
            findings: List of finding dictionaries to import

        Returns:
            Tuple of (failed_count, success_count, failed_findings)
        """
        failed_count = 0
        success_count = 0
        failed_findings = []

        for finding_data in findings:
            try:
                if (
                    not isinstance(finding_data["Resources"], list)
                    or len(finding_data["Resources"]) == 0
                ):
                    raise InvalidInputException(
                        op="BatchImportFindings",
                        msg="Finding must contain at least one resource in the Resources array",
                    )

                finding_id = finding_data["Id"]

                existing_finding = next(
                    (f for f in self.findings if f.id == finding_id), None
                )

                if existing_finding:
                    existing_finding.data.update(finding_data)
                else:
                    new_finding = Finding(finding_id, finding_data)
                    self.findings.append(new_finding)

                success_count += 1

            except Exception as e:
                failed_count += 1
                failed_findings.append(
                    {
                        "Id": finding_data.get("Id", ""),
                        "ErrorCode": "InvalidInput",
                        "ErrorMessage": str(e),
                    }
                )

        return failed_count, success_count, failed_findings

    def enable_organization_admin_account(self, admin_account_id: str) -> None:
        try:
            org = self.org_backend.describe_organization()
            org_id = org["Organization"]["Id"]
        except RESTError:
            raise AWSOrganizationsNotInUseException()

        if self.account_id != org["Organization"]["MasterAccountId"]:
            raise RESTError(
                "AccessDeniedException",
                "The request was rejected because you don't have sufficient permissions "
                "to perform this operation. The security token included in the request "
                "is for an account that isn't authorized to perform this operation.",
            )

        try:
            self.org_backend.get_account_by_id(admin_account_id)
        except RESTError:
            raise RESTError(
                "InvalidInputException",
                f"The request was rejected because the account {admin_account_id} is not "
                f"a member of organization {org_id}.",
            )

        org_config = self._get_org_config()
        org_config["admin_account_id"] = admin_account_id

    def update_organization_configuration(
        self,
        auto_enable: bool,
        auto_enable_standards: Optional[str] = None,
        organization_configuration: Optional[Dict[str, Any]] = None,
    ) -> None:
        try:
            self.org_backend.describe_organization()
        except RESTError:
            raise RESTError(
                "ResourceNotFoundException",
                "The request was rejected because AWS Organizations is not in use or not "
                "configured for this account.",
            )

        org_config = self._get_org_config()
        if not org_config["admin_account_id"]:
            raise RESTError(
                "ResourceNotFoundException",
                "The request was rejected because no administrator account has been designated.",
            )

        if self.account_id != org_config["admin_account_id"]:
            raise RESTError(
                "AccessDeniedException",
                "The request was rejected because you don't have permission to perform "
                "this action. Only the designated administrator account can update the "
                "organization configuration.",
            )

        if organization_configuration:
            config_type = organization_configuration.get("ConfigurationType")
            if config_type not in ["CENTRAL", "LOCAL"]:
                raise RESTError(
                    "InvalidInputException",
                    "The request was rejected because the ConfigurationType value must be "
                    "either CENTRAL or LOCAL.",
                )

            status = organization_configuration.get("Status")
            if status not in ["PENDING", "ENABLED", "FAILED"]:
                raise RESTError(
                    "InvalidInputException",
                    "The request was rejected because the Status value must be one of "
                    "PENDING, ENABLED, or FAILED.",
                )

            if config_type == "CENTRAL":
                if auto_enable:
                    raise RESTError(
                        "ValidationException",
                        "The request was rejected because AutoEnable must be false when "
                        "ConfigurationType is CENTRAL.",
                    )
                if auto_enable_standards != "NONE":
                    raise RESTError(
                        "ValidationException",
                        "The request was rejected because AutoEnableStandards must be NONE "
                        "when ConfigurationType is CENTRAL.",
                    )

            org_config["configuration"] = organization_configuration

        org_config["auto_enable"] = auto_enable

        if auto_enable_standards is not None:
            if auto_enable_standards not in ["NONE", "DEFAULT"]:
                raise RESTError(
                    "InvalidInputException",
                    "The request was rejected because AutoEnableStandards must be either "
                    "NONE or DEFAULT.",
                )
            org_config["auto_enable_standards"] = auto_enable_standards

    def get_administrator_account(self) -> Dict[str, Any]:
        try:
            org = self.org_backend.describe_organization()
            management_account_id = org["Organization"]["MasterAccountId"]
        except RESTError:
            return {}

        org_config = self._get_org_config()
        admin_account_id = org_config["admin_account_id"]

        if not admin_account_id:
            return {}

        if (
            self.account_id == management_account_id
            or self.account_id == admin_account_id
        ):
            return {}

        return {
            "Administrator": {
                "AccountId": admin_account_id,
                "MemberStatus": "ENABLED",
                "InvitationId": f"invitation-{admin_account_id}",
                "InvitedAt": datetime.datetime.now().isoformat(),
            }
        }

    def describe_organization_configuration(self) -> Dict[str, Any]:
        try:
            self.org_backend.describe_organization()
        except RESTError:
            raise RESTError(
                "AccessDeniedException",
                "You do not have sufficient access to perform this action.",
            )

        org_config = self._get_org_config()
        if not org_config["admin_account_id"]:
            raise RESTError(
                "AccessDeniedException",
                "You do not have sufficient access to perform this action.",
            )

        if self.account_id != org_config["admin_account_id"]:
            raise RESTError(
                "AccessDeniedException",
                "You do not have sufficient access to perform this action.",
            )

        return {
            "AutoEnable": org_config["auto_enable"],
            "MemberAccountLimitReached": False,
            "AutoEnableStandards": org_config["auto_enable_standards"],
            "OrganizationConfiguration": org_config["configuration"],
        }


securityhub_backends = BackendDict(SecurityHubBackend, "securityhub")



import os
from tempfile import TemporaryFile

import boto3
import pytest

from . import s3_aws_verified


@pytest.mark.aws_verified
@s3_aws_verified
@pytest.mark.parametrize("encoding", [None, "gzip", "unknown"])
def test_content_encoding_is_returned(encoding, bucket_name=None):
    client = boto3.client("s3", region_name="us-east-1")
    s3_file = "some.file"

    attrs = {
        "ContentType": "application/octet-stream",
        "ACL": "bucket-owner-full-control",
    }
    if encoding:
        attrs["ContentEncoding"] = encoding

    with TemporaryFile() as f:
        f.write(os.urandom(1024))
        f.flush()
        client.upload_fileobj(f, bucket_name, s3_file, ExtraArgs=attrs)

    res = client.get_object(Bucket=bucket_name, Key=s3_file)
    assert res.get("ContentEncoding") == encoding



import json

import boto3
import requests

from moto import mock_aws, settings
from tests import DEFAULT_ACCOUNT_ID

from .test_ecr_helpers import _create_image_manifest

example_response = {
    "imageScanFindings": {
        "enhancedFindings": [
            {
                "awsAccountId": DEFAULT_ACCOUNT_ID,
                "description": "When curl is asked to use HSTS, the expiry time for a subdomain might overwrite a parent domain's cache entry, making it end sooner or later than otherwise intended. This affects curl using applications that enable HSTS and use URLs with the insecure `HTTP://` scheme and perform transfers with hosts like `x.example.com` as well as `example.com` where the first host is a subdomain of the second host. (The HSTS cache either needs to have been populated manually or there needs to have been previous HTTPS accesses done as the cache needs to have entries for the domains involved to trigger this problem.) When `x.example.com` responds with `Strict-Transport-Security:` headers, this bug can make the subdomain's expiry timeout *bleed over* and get set for the parent domain `example.com` in curl's HSTS cache. The result of a triggered bug is that HTTP accesses to `example.com` get converted to HTTPS for a different period of time than what was asked for by the origin server. If `example.com` for example stops supporti",
                "findingArn": "arn:aws:inspector2:eu-west-1:000000000000:finding/92ec736dde6c50c668cf5721027ae176",
                "packageVulnerabilityDetails": {
                    "cvss": [
                        {
                            "baseScore": 6.5,
                            "scoringVector": "CVSS:3.1/AV:N/AC:H/PR:N/UI:N/S:U/C:N/I:H/A:L",
                            "source": "NVD",
                            "version": "3.1",
                        }
                    ],
                    "referenceUrls": [
                        "https://bugs.debian.org/cgi-bin/bugreport.cgi?bug=1086804"
                    ],
                    "relatedVulnerabilities": [],
                    "source": "DEBIAN_CVE",
                    "sourceUrl": "https://security-tracker.debian.org/tracker/CVE-2024-9681",
                    "vendorSeverity": "not yet assigned",
                    "vulnerabilityId": "CVE-2024-9681",
                    "vulnerablePackages": [
                        {
                            "arch": "AMD64",
                            "epoch": 0,
                            "name": "curl",
                            "packageManager": "OS",
                            "release": "10+deb12u8",
                            "sourceLayerHash": "sha256:416e17e4fbb3912125a75f472458c684de6c08405c5c0d2221649b7feac4f78c",
                            "version": "7.88.1",
                            "fixedInVersion": "NotAvailable",
                        }
                    ],
                },
                "remediation": {"recommendation": {"text": "None Provided"}},
                "resources": [
                    {
                        "details": {
                            "awsEcrContainerImage": {
                                "architecture": "amd64",
                                "imageHash": "sha256:9c7d02b8b260fa42b14a85cc1221f942f2029b5bf1132fc300ec479273406f5d",
                                "imageTags": ["5.0.28"],
                                "platform": "DEBIAN_12",
                                "registry": DEFAULT_ACCOUNT_ID,
                                "repositoryName": "prod/cp/motoserver",
                            }
                        },
                        "id": "arn:aws:ecr:eu-west-1:000000000000:repository/prod/cp/motoserver/sha256:9c7d02b8b260fa42b14a85cc1221f942f2029b5bf1132fc300ec479273406f5d",
                        "tags": {},
                        "type": "AWS_ECR_CONTAINER_IMAGE",
                    }
                ],
                "score": 6.5,
                "scoreDetails": {
                    "cvss": {
                        "adjustments": [],
                        "score": 6.5,
                        "scoreSource": "NVD",
                        "scoringVector": "CVSS:3.1/AV:N/AC:H/PR:N/UI:N/S:U/C:N/I:H/A:L",
                        "version": "3.1",
                    }
                },
                "severity": "MEDIUM",
                "status": "ACTIVE",
                "title": "CVE-2024-9681 - curl",
                "type": "PACKAGE_VULNERABILITY",
                "fixAvailable": "NO",
                "exploitAvailable": "NO",
            },
            {
                "awsAccountId": DEFAULT_ACCOUNT_ID,
                "description": "None Provided",
                "findingArn": "arn:aws:inspector2:eu-west-1:000000000000:finding/9ff1469a4702529c61a8f00661a27576",
                "packageVulnerabilityDetails": {
                    "cvss": [],
                    "referenceUrls": [
                        "https://bugs.debian.org/cgi-bin/bugreport.cgi?bug=1094730"
                    ],
                    "relatedVulnerabilities": [],
                    "source": "DEBIAN_CVE",
                    "sourceUrl": "https://security-tracker.debian.org/tracker/CVE-2025-24528",
                    "vendorSeverity": "not yet assigned",
                    "vulnerabilityId": "CVE-2025-24528",
                    "vulnerablePackages": [
                        {
                            "arch": "AMD64",
                            "epoch": 0,
                            "name": "krb5",
                            "packageManager": "OS",
                            "release": "2+deb12u2",
                            "sourceLayerHash": "sha256:951ef04389205caf59aa51717bdd03e0765e97a42bd488de11ebe7b05db2b525",
                            "version": "1.20.1",
                            "fixedInVersion": "NotAvailable",
                        }
                    ],
                },
                "remediation": {"recommendation": {"text": "None Provided"}},
                "resources": [
                    {
                        "details": {
                            "awsEcrContainerImage": {
                                "architecture": "amd64",
                                "imageHash": "sha256:9c7d02b8b260fa42b14a85cc1221f942f2029b5bf1132fc300ec479273406f5d",
                                "imageTags": ["5.0.28"],
                                "platform": "DEBIAN_12",
                                "registry": DEFAULT_ACCOUNT_ID,
                                "repositoryName": "prod/cp/motoserver",
                            }
                        },
                        "id": "arn:aws:ecr:eu-west-1:000000000000:repository/prod/cp/motoserver/sha256:9c7d02b8b260fa42b14a85cc1221f942f2029b5bf1132fc300ec479273406f5d",
                        "tags": {},
                        "type": "AWS_ECR_CONTAINER_IMAGE",
                    }
                ],
                "score": 0.0,
                "severity": "UNTRIAGED",
                "status": "ACTIVE",
                "title": "CVE-2025-24528 - krb5",
                "type": "PACKAGE_VULNERABILITY",
                "fixAvailable": "NO",
                "exploitAvailable": "NO",
            },
        ],
        "findingSeverityCounts": {"MEDIUM": 1, "UNTRIAGED": 1},
    },
    "registryId": DEFAULT_ACCOUNT_ID,
    "repositoryName": "reponame",
    "imageId": {"imageTag": "latest"},
    "imageScanStatus": {
        "status": "COMPLETE",
        "description": "The scan was completed successfully.",
    },
}


@mock_aws
def test_set_findings():
    base_url = (
        "localhost:5000" if settings.TEST_SERVER_MODE else "motoapi.amazonaws.com"
    )

    findings = {
        "results": [example_response],
        "region": "us-west-1",
    }
    resp = requests.post(
        f"http://{base_url}/moto-api/static/ecr/scan-finding-results",
        json=findings,
    )
    assert resp.status_code == 201

    ecr = boto3.client("ecr", region_name="us-west-1")
    ecr.create_repository(repositoryName="reponame")
    image_tag = "latest"
    image_digest = ecr.put_image(
        repositoryName="reponame",
        imageManifest=json.dumps(_create_image_manifest()),
        imageTag=image_tag,
    )["image"]["imageId"]["imageDigest"]

    # Diversion
    ecr.create_repository(repositoryName="reponame2")
    ecr.put_image(
        repositoryName="reponame2",
        imageManifest=json.dumps(_create_image_manifest()),
        imageTag="latest",
    )
    ecr.start_image_scan(repositoryName="reponame2", imageId={"imageTag": "latest"})

    # when
    ecr.start_image_scan(repositoryName="reponame", imageId={"imageTag": image_tag})

    findings = ecr.describe_image_scan_findings(
        repositoryName="reponame", imageId={"imageTag": image_tag}
    )
    findings.pop("ResponseMetadata")
    assert findings == example_response

    # Retrieving results for different repo returns the default response
    diff_findings = ecr.describe_image_scan_findings(
        repositoryName="reponame2", imageId={"imageTag": image_tag}
    )
    assert diff_findings["imageScanFindings"]["findingSeverityCounts"]["HIGH"] == 1

    # Retrieving results using imageDigest also returns the default (because it's not configured)
    diff_findings = ecr.describe_image_scan_findings(
        repositoryName="reponame", imageId={"imageDigest": image_digest}
    )
    assert diff_findings["imageScanFindings"]["findingSeverityCounts"]["HIGH"] == 1

    # Calling list_findings with original arguments returns original list
    findings = ecr.describe_image_scan_findings(
        repositoryName="reponame", imageId={"imageTag": image_tag}
    )
    findings.pop("ResponseMetadata")
    assert findings == example_response



import datetime

import boto3
import botocore.exceptions
import pytest
from botocore.client import ClientError

from moto import mock_aws
from moto.core.utils import utcnow
from moto.s3.responses import DEFAULT_REGION_NAME

from . import s3_aws_verified
from .test_s3 import enable_versioning


@mock_aws
def test_head_object_if_modified_since():
    s3_client = boto3.client("s3", region_name=DEFAULT_REGION_NAME)
    bucket_name = "blah"
    s3_client.create_bucket(Bucket=bucket_name)

    key = "hello.txt"

    s3_client.put_object(Bucket=bucket_name, Key=key, Body="test")

    with pytest.raises(botocore.exceptions.ClientError) as err:
        s3_client.head_object(
            Bucket=bucket_name,
            Key=key,
            IfModifiedSince=utcnow() + datetime.timedelta(hours=1),
        )
    err_value = err.value
    assert err_value.response["Error"] == {"Code": "304", "Message": "Not Modified"}


@mock_aws
def test_head_object_if_modified_since_refresh():
    s3_client = boto3.client("s3", region_name=DEFAULT_REGION_NAME)
    bucket_name = "blah"
    s3_client.create_bucket(Bucket=bucket_name)

    key = "hello.txt"

    s3_client.put_object(Bucket=bucket_name, Key=key, Body="test")

    response = s3_client.head_object(Bucket=bucket_name, Key=key)

    with pytest.raises(botocore.exceptions.ClientError) as err:
        s3_client.head_object(
            Bucket=bucket_name,
            Key=key,
            IfModifiedSince=response["LastModified"],
        )
    err_value = err.value
    assert err_value.response["Error"] == {"Code": "304", "Message": "Not Modified"}


@mock_aws
def test_head_object_if_unmodified_since():
    s3_client = boto3.client("s3", region_name=DEFAULT_REGION_NAME)
    bucket_name = "blah"
    s3_client.create_bucket(Bucket=bucket_name)

    key = "hello.txt"

    s3_client.put_object(Bucket=bucket_name, Key=key, Body="test")

    with pytest.raises(botocore.exceptions.ClientError) as err:
        s3_client.head_object(
            Bucket=bucket_name,
            Key=key,
            IfUnmodifiedSince=utcnow() - datetime.timedelta(hours=1),
        )
    err_value = err.value
    assert err_value.response["Error"] == {
        "Code": "412",
        "Message": "Precondition Failed",
    }


@mock_aws
def test_head_object_if_match():
    s3_client = boto3.client("s3", region_name=DEFAULT_REGION_NAME)
    bucket_name = "blah"
    s3_client.create_bucket(Bucket=bucket_name)

    key = "hello.txt"

    s3_client.put_object(Bucket=bucket_name, Key=key, Body="test")

    with pytest.raises(botocore.exceptions.ClientError) as err:
        s3_client.head_object(Bucket=bucket_name, Key=key, IfMatch='"hello"')
    err_value = err.value
    assert err_value.response["Error"] == {
        "Code": "412",
        "Message": "Precondition Failed",
    }


@mock_aws
def test_head_object_if_none_match():
    s3_client = boto3.client("s3", region_name=DEFAULT_REGION_NAME)
    bucket_name = "blah"
    s3_client.create_bucket(Bucket=bucket_name)

    key = "hello.txt"

    etag = s3_client.put_object(Bucket=bucket_name, Key=key, Body="test")["ETag"]

    with pytest.raises(botocore.exceptions.ClientError) as err:
        s3_client.head_object(Bucket=bucket_name, Key=key, IfNoneMatch=etag)
    err_value = err.value
    assert err_value.response["Error"] == {"Code": "304", "Message": "Not Modified"}
    assert err_value.response["ResponseMetadata"]["HTTPHeaders"]["etag"] == etag


@mock_aws
def test_get_object_if_modified_since_refresh():
    s3_client = boto3.client("s3", region_name=DEFAULT_REGION_NAME)
    bucket_name = "blah"
    s3_client.create_bucket(Bucket=bucket_name)

    key = "hello.txt"

    s3_client.put_object(Bucket=bucket_name, Key=key, Body="test")

    response = s3_client.get_object(Bucket=bucket_name, Key=key)

    with pytest.raises(botocore.exceptions.ClientError) as err:
        s3_client.get_object(
            Bucket=bucket_name,
            Key=key,
            IfModifiedSince=response["LastModified"],
        )
    err_value = err.value
    assert err_value.response["Error"] == {"Code": "304", "Message": "Not Modified"}


@mock_aws
def test_get_object_if_modified_since():
    s3_client = boto3.client("s3", region_name=DEFAULT_REGION_NAME)
    bucket_name = "blah"
    s3_client.create_bucket(Bucket=bucket_name)

    key = "hello.txt"

    s3_client.put_object(Bucket=bucket_name, Key=key, Body="test")

    with pytest.raises(botocore.exceptions.ClientError) as err:
        s3_client.get_object(
            Bucket=bucket_name,
            Key=key,
            IfModifiedSince=utcnow() + datetime.timedelta(hours=1),
        )
    err_value = err.value
    assert err_value.response["Error"] == {"Code": "304", "Message": "Not Modified"}


@mock_aws
def test_get_object_if_unmodified_since():
    s3_client = boto3.client("s3", region_name=DEFAULT_REGION_NAME)
    bucket_name = "blah"
    s3_client.create_bucket(Bucket=bucket_name)

    key = "hello.txt"

    s3_client.put_object(Bucket=bucket_name, Key=key, Body="test")

    with pytest.raises(botocore.exceptions.ClientError) as err:
        s3_client.get_object(
            Bucket=bucket_name,
            Key=key,
            IfUnmodifiedSince=utcnow() - datetime.timedelta(hours=1),
        )
    err_value = err.value
    assert err_value.response["Error"]["Code"] == "PreconditionFailed"
    assert err_value.response["Error"]["Condition"] == "If-Unmodified-Since"


@mock_aws
def test_get_object_if_match():
    s3_client = boto3.client("s3", region_name=DEFAULT_REGION_NAME)
    bucket_name = "blah"
    s3_client.create_bucket(Bucket=bucket_name)

    key = "hello.txt"

    s3_client.put_object(Bucket=bucket_name, Key=key, Body="test")

    with pytest.raises(botocore.exceptions.ClientError) as err:
        s3_client.get_object(Bucket=bucket_name, Key=key, IfMatch='"hello"')
    err_value = err.value
    assert err_value.response["Error"]["Code"] == "PreconditionFailed"
    assert err_value.response["Error"]["Condition"] == "If-Match"


@mock_aws
def test_get_object_if_none_match():
    s3_client = boto3.client("s3", region_name=DEFAULT_REGION_NAME)
    bucket_name = "blah"
    s3_client.create_bucket(Bucket=bucket_name)

    key = "hello.txt"

    etag = s3_client.put_object(Bucket=bucket_name, Key=key, Body="test")["ETag"]

    with pytest.raises(botocore.exceptions.ClientError) as err:
        s3_client.get_object(Bucket=bucket_name, Key=key, IfNoneMatch=etag)
    err_value = err.value
    assert err_value.response["Error"] == {"Code": "304", "Message": "Not Modified"}
    assert err_value.response["ResponseMetadata"]["HTTPHeaders"]["etag"] == etag


@s3_aws_verified
@pytest.mark.aws_verified
@pytest.mark.parametrize("versioned", [True, False], ids=["versioned", "not versioned"])
def test_put_object_if_none_match(versioned, bucket_name=None):
    s3 = boto3.client("s3", region_name="us-east-1")
    if versioned:
        enable_versioning(bucket_name, s3)

    s3.put_object(
        Key="test_object",
        Body="test",
        Bucket=bucket_name,
        IfNoneMatch="*",
    )

    with pytest.raises(ClientError) as exc:
        s3.put_object(
            Key="test_object",
            Body="another_test",
            Bucket=bucket_name,
            IfNoneMatch="*",
        )
    err = exc.value.response["Error"]
    assert err["Code"] == "PreconditionFailed"
    assert (
        err["Message"]
        == "At least one of the pre-conditions you specified did not hold"
    )
    assert err["Condition"] == "If-None-Match"


@s3_aws_verified
@pytest.mark.aws_verified
def test_put_object_if_match(bucket_name=None):
    s3 = boto3.client("s3", region_name="us-east-1")
    etag = s3.put_object(Bucket=bucket_name, Key="test_key", Body=b"test")["ETag"]

    with pytest.raises(ClientError) as exc:
        s3.put_object(Bucket=bucket_name, Key="test_key", Body=b"test2", IfMatch="test")
    err = exc.value.response["Error"]
    assert err["Code"] == "PreconditionFailed"
    assert (
        err["Message"]
        == "At least one of the pre-conditions you specified did not hold"
    )
    assert err["Condition"] == "If-Match"

    # We can if we match the etag
    etag2 = s3.put_object(
        Bucket=bucket_name, Key="test_key", Body=b"test2", IfMatch=etag
    )["ETag"]
    assert etag != etag2

    # S3 is flexible - it will accept ETags without quotes
    etag_without_quotes = etag2.replace('"', "")
    etag3 = s3.put_object(
        Bucket=bucket_name, Key="test_key", Body=b"test3", IfMatch=etag_without_quotes
    )["ETag"]

    # And ETags with one quote removed
    s3.put_object(Bucket=bucket_name, Key="test_key", Body=b"test4", IfMatch=etag3[1:])

    # It will fail when specifying an unknown key though
    with pytest.raises(ClientError) as exc:
        s3.put_object(Bucket=bucket_name, Key="unknown", Body=b"test", IfMatch="test")
    err = exc.value.response["Error"]
    assert err["Code"] == "NoSuchKey"
